import autonetkit.ank as ank_utils
import autonetkit.log as log


def build_layer1(anm):
    g_l1 = anm.add_overlay('layer1')
    g_phy = anm['phy']
    g_l1.add_nodes_from(g_phy)
    g_l1.add_edges_from(g_phy.edges())

    # aggregate collision domains
    hubs = g_l1.nodes(device_type="hub")
    ank_utils.aggregate_nodes(g_l1, hubs)
    # TODO: remove aggregated and disconnected nodes
    # refresh hubs list
    hubs = g_l1.nodes(device_type="hub")

    for hub in hubs:
        hub.collision_domain = True

    split_ptp(anm)
    build_layer1_conn(anm)

def build_layer1_conn(anm):
    g_l1 = anm['layer1']
    g_l1_conn = anm.add_overlay('layer1_conn')
    g_l1_conn.add_nodes_from(g_l1, retain="collision_domain")
    g_l1_conn.add_edges_from(g_l1.edges())

    collision_domains = g_l1_conn.nodes(collision_domain=True)
    exploded_edges = ank_utils.explode_nodes(g_l1_conn, collision_domains)

    # explode each seperately?
    for edge in exploded_edges:
        edge.multipoint = True
        edge.src_int.multipoint = True
        edge.dst_int.multipoint = True

    # TODO: tidy up partial repetition of collision_domain attribute and
    # device type


def split_ptp(anm):
    g_l1 = anm['layer1']
    g_phy = anm['phy']

    edges_to_split = [e for e in g_l1.edges()]
    edges_to_split = [e for e in edges_to_split
                      if not (e.src.is_hub() or e.dst.is_hub())]

    # TODO: debug the edges to split
    # print "edges to split", edges_to_split
    for edge in edges_to_split:
        edge.split = True  # mark as split for use in building nidb

    split_created_nodes = list(ank_utils.split(g_l1, edges_to_split,
                                               retain=['split'],
                                               id_prepend='cd_'))

    for node in split_created_nodes:
        node.device_type = "collision_domain"
        node.collision_domain = True

    # TODO: if parallel nodes, offset
    # TODO: remove graphics, assign directly
    g_graphics = anm['graphics']

    if len(g_graphics):
        co_ords_overlay = g_graphics  # source from graphics overlay
    else:
        co_ords_overlay = g_phy  # source from phy overlay

    for node in split_created_nodes:
        node['graphics'].x = ank_utils.neigh_average(g_l1, node, 'x',
                                                     co_ords_overlay) + 0.1

        # temporary fix for gh-90

        node['graphics'].y = ank_utils.neigh_average(g_l1, node, 'y',
                                                     co_ords_overlay) + 0.1

        # temporary fix for gh-90

        asn = ank_utils.neigh_most_frequent(
            g_l1, node, 'asn', g_phy)  # arbitrary choice
        node['graphics'].asn = asn
        node.asn = asn  # need to use asn in IP overlay for aggregating subnets


# TODO: build layer 1 connectivity graph



def house():
    """Returns anm with  input and physical as house graph"""
    from networkx.readwrite import json_graph
    import networkx as nx
    import autonetkit
    # returns a house graph
    data = {'directed': False,
    'graph': [],
    'links': [{'_ports': {'r4': 2, 'r5': 1},
    'raw_interfaces': {},
    'source': 0,
    'target': 1},
    {'_ports': {'r2': 3, 'r4': 1},
    'raw_interfaces': {},
    'source': 0,
    'target': 3},
    {'_ports': {'r3': 3, 'r5': 2},
    'raw_interfaces': {},
    'source': 1,
    'target': 4},
    {'_ports': {'r1': 1, 'r2': 1},
    'raw_interfaces': {},
    'source': 2,
    'target': 3},
    {'_ports': {'r1': 2, 'r3': 1},
    'raw_interfaces': {},
    'source': 2,
    'target': 4},
    {'_ports': {'r2': 2, 'r3': 2},
    'raw_interfaces': {},
    'source': 3,
    'target': 4}],
    'multigraph': False,
    'nodes': [{'_ports': {0: {'category': 'physical', 'description': None},
    1: {'category': 'physical', 'description': 'r4 to r2', 'id': 'eth0'},
    2: {'category': 'physical', 'description': 'r4 to r5', 'id': 'eth1'}},
    'asn': 2,
    'device_type': 'router',
    'id': 'r4',
    'label': 'r4',
    'x': 675,
    'y': 300},
    {'_ports': {0: {'category': 'physical', 'description': None},
    1: {'category': 'physical', 'description': 'r5 to r4', 'id': 'eth0'},
    2: {'category': 'physical', 'description': 'r5 to r3', 'id': 'eth1'}},
    'asn': 2,
    'device_type': 'router',
    'id': 'r5',
    'label': 'r5',
    'x': 675,
    'y': 500},
    {'_ports': {0: {'category': 'physical', 'description': None},
    1: {'category': 'physical', 'description': 'r1 to r2', 'id': 'eth0'},
    2: {'category': 'physical', 'description': 'r1 to r3', 'id': 'eth1'}},
    'asn': 1,
    'device_type': 'router',
    'id': 'r1',
    'label': 'r1',
    'x': 350,
    'y': 400},
    {'_ports': {0: {'category': 'physical', 'description': None},
    1: {'category': 'physical', 'description': 'r2 to r1', 'id': 'eth0'},
    2: {'category': 'physical', 'description': 'r2 to r3', 'id': 'eth1'},
    3: {'category': 'physical', 'description': 'r2 to r4', 'id': 'eth2'}},
    'asn': 1,
    'device_type': 'router',
    'id': 'r2',
    'label': 'r2',
    'x': 500,
    'y': 300},
    {'_ports': {0: {'category': 'physical', 'description': None},
    1: {'category': 'physical', 'description': 'r3 to r1', 'id': 'eth0'},
    2: {'category': 'physical', 'description': 'r3 to r2', 'id': 'eth1'},
    3: {'category': 'physical', 'description': 'r3 to r5', 'id': 'eth2'}},
    'asn': 1,
    'device_type': 'router',
    'id': 'r3',
    'label': 'r3',
    'x': 500,
    'y': 500}]}
    graph = json_graph.node_link_graph(data)
    anm = autonetkit.anm.NetworkModel()
    g_in = anm.add_overlay("input")
    g_in._replace_graph(nx.Graph(graph))
    #TODO: check if should build overlays here rather than clone in?
    g_phy = anm["phy"]
    g_phy._replace_graph(graph)
    return anm


def initialize(filename, defaults = True):
    from autonetkit.build_network import load, initialise
    with open(filename) as fh:
        data = fh.read()

    graph = load(data, defaults=defaults)
    anm = initialise(graph)
    return anm


from collections import OrderedDict

import autonetkit.log as log

# based on docs.python.org/2.7/library/collections
# and http://stackoverflow.com/q/455059


class ConfigStanza(object):

    def __init__(self, *args, **kwargs):
        if len(args) == 1 and isinstance(args[0], ConfigStanza):
            # Clone the data (shallow copy)
            # TODO: check how this relates to calling dict() on a dict - same?
            in_dict = args[0]._odict
            object.__setattr__(self, '_odict', OrderedDict(in_dict))
            return

        if len(args) == 1 and isinstance(args[0], dict):
            # Clone the data (shallow copy)
            in_dict = args[0]
            object.__setattr__(self, '_odict', OrderedDict(in_dict))
            return

        object.__setattr__(self, '_odict', OrderedDict(kwargs))

    def __repr__(self):
        return str(self._odict.items())

    def to_json(self):
        retval = OrderedDict(self._odict)  # clone to append to
        retval['_ConfigStanza'] = True
        return retval

    def add_stanza(self, name, **kwargs):
        """Adds a sub-stanza to this stanza"""
        stanza = ConfigStanza(**kwargs)
        self[name] = stanza
        return stanza

    def __getitem__(self, key):
        return self._odict[key]

    def __delitem__(self, key):
        del self._odict[key]

    def __setitem__(self, key, value):
        if isinstance(value, dict):
            log.warning(
                "Adding dictionary %s: did you mean ",
                "to add a ConfigStanza?" % key)
        self._odict[key] = value

    def __setattr__(self, key, value):
        self._odict[key] = value

    def __getattr__(self, key):
        # TODO: decide how to return misses
        try:
            return self._odict[key]
        except KeyError:
            # TODO: implement warning here
            # TODO: log a key miss, and if strict turned on, give warning
            return

    def items(self):
        # TODO map this to proper dict inherit to
        # support these methods, keys, etc
        return self._odict.items()

    def __iter__(self):
        return iter(self._odict.items())

    def __len__(self):
        return len(self._odict)

    # TODO: add __iter__, __keys etc

    #self.__dict__['_odict'][key] = value

    # TODO: add a sort function that sorts the OrderedDict



"""Module to build overlay graphs for network design"""

import autonetkit
import autonetkit.ank as ank_utils
import autonetkit.anm
import autonetkit.config
import autonetkit.exception
import autonetkit.log as log
import networkx as nx

SETTINGS = autonetkit.config.settings

# TODO: revisit phy_neighbors for eg ASN and use layer3 instead

__all__ = ['build']


def load(input_graph_string, defaults = True):

    # TODO: look at XML header for file type
    import autonetkit.load.graphml as graphml
    import autonetkit.load.load_json as load_json
    try:
        input_graph = graphml.load_graphml(input_graph_string, defaults=defaults)
    except autonetkit.exception.AnkIncorrectFileFormat:
        try:
            input_graph = load_json.load_json(input_graph_string, defaults=defaults)
        except (ValueError, autonetkit.exception.AnkIncorrectFileFormat):
            # try a different reader
            try:
                from autonetkit_cisco import load as cisco_load
            except ImportError, error:
                log.debug("Unable to load autonetkit_cisco %s", error)
                return  # module not present (development module)
            else:
                input_graph = cisco_load.load(input_graph_string)
                # add local deployment host
                SETTINGS['General']['deploy'] = True
                SETTINGS['Deploy Hosts']['internal'] = {
                    'VIRL': {
                        'deploy': True,
                    },
                }

    return input_graph


def grid_2d(dim):
    """Creates a 2d grid of dimension dim"""
    graph = nx.grid_2d_graph(dim, dim)

    for node in graph:
        graph.node[node]['asn'] = 1
        graph.node[node]['x'] = node[0] * 150
        graph.node[node]['y'] = node[1] * 150
        graph.node[node]['device_type'] = 'router'
        graph.node[node]['platform'] = 'cisco'
        graph.node[node]['syntax'] = 'ios_xr'
        graph.node[node]['host'] = 'internal'
        graph.node[node]['ibgp_role'] = "Peer"

    mapping = {node: "%s_%s" % (node[0], node[1]) for node in graph}
    # Networkx wipes data if remap with same labels
    nx.relabel_nodes(graph, mapping, copy=False)
    for src, dst in graph.edges():
        graph[src][dst]['type'] = "physical"
        # add global index for sorting

    SETTINGS['General']['deploy'] = True
    SETTINGS['Deploy Hosts']['internal'] = {
        'cisco': {
            'deploy': True,
        },
    }

    return graph


def initialise(input_graph):
    """Initialises the input graph with from a NetworkX graph"""
    all_multigraph = input_graph.is_multigraph()
    anm = autonetkit.anm.NetworkModel(all_multigraph=all_multigraph)

    g_in = anm.initialise_input(input_graph)
    # autonetkit.update_vis(anm)

# set defaults
    if not g_in.data.specified_int_names:
        # if not specified then automatically assign interface names
        g_in.data.specified_int_names = False

    #import autonetkit.plugins.graph_product as graph_product
    # graph_product.expand(g_in)  # apply graph products if relevant

    expand_fqdn = False
    # TODO: make this set from config and also in the input file
    if expand_fqdn and len(ank_utils.unique_attr(g_in, "asn")) > 1:
        # Multiple ASNs set, use label format device.asn
        anm.set_node_label(".", ['label', 'asn'])

    g_in.update(g_in.routers(platform="junosphere"), syntax="junos")
    g_in.update(g_in.routers(platform="dynagen"), syntax="ios")
    g_in.update(g_in.routers(platform="netkit"), syntax="quagga")
    # TODO: is this used?
    g_in.update(g_in.servers(platform="netkit"), syntax="quagga")

    #TODO: check this is needed
    #autonetkit.ank.set_node_default(g_in, specified_int_names=None)

    g_graphics = anm.add_overlay("graphics")  # plotting data
    g_graphics.add_nodes_from(g_in, retain=['x', 'y', 'device_type',
                                            'label', 'device_subtype', 'asn'])

    return anm


def check_server_asns(anm):
    """Checks that servers have appropriate ASN allocated.
    Warns and auto-corrects servers connected to routers of a different AS
    #TODO: provide manual over-ride for this auto-correct.
    """
    # TODO: Move to validate module?
    g_phy = anm['phy']

    for server in g_phy.servers():
        #TODO: remove now have external_connector device_type?
        if server.device_subtype in ("SNAT", "FLAT"):
            continue  # Don't warn on ASN for NAT elements
        l3_neighbors = list(server['layer3'].neighbors())
        l3_neighbor_asns = set(n.asn for n in l3_neighbors)
        if server.asn not in l3_neighbor_asns:
            neighs_with_asn = ["%s: AS %s" % (n, n.asn)
                               for n in l3_neighbors]
            # tuples for warning message
            server.log.warning("Server does not belong to same ASN "
                               "as neighbors %s" % (neighs_with_asn))

            if len(l3_neighbors) == 1:
                # single ASN of neighbor -> auto correct
                if server['input'].default_asn:
                    neigh_asn = l3_neighbor_asns.pop()
                    log.warning("Updating server %s AS from %s"
                                " to %s", server, server.asn, neigh_asn)
                    server.asn = neigh_asn
                else:
                    log.info("Server %s ASN %s explictly set by user, "
                             "not auto-correcting", server, server.asn)


def apply_design_rules(anm):
    """Applies appropriate design rules to ANM"""
    # log.info("Building overlay topologies")
    g_in = anm['input']

    build_phy(anm)

    try:
        from autonetkit_cisco import build_network as cisco_build_network
    except ImportError, e:
        log.debug("Unable to load autonetkit_cisco %s", e)
    else:
        cisco_build_network.post_phy(anm)

    g_phy = anm['phy']
    from autonetkit.design.osi_layers import build_layer1, build_layer2, build_layer3
    # log.info("Building layer2")
    build_layer1(anm)
    build_layer2(anm)
    # autonetkit.update_http(anm)

    # log.info("Building layer3")
    build_layer3(anm)

    check_server_asns(anm)

    from autonetkit.design.mpls import build_vrf
    build_vrf(anm)  # do before to add loopbacks before ip allocations
    from autonetkit.design.ip import build_ip, build_ipv4, build_ipv6
    # TODO: replace this with layer2 overlay topology creation
    # log.info("Allocating IP addresses")
    build_ip(anm)  # ip infrastructure topology

    address_family = g_in.data.address_family or "v4"  # default is v4
# TODO: can remove the infrastructure now create g_ip seperately
    if address_family == "None":
        log.info("IP addressing disabled, disabling routing protocol ",
                 "configuration")
        anm['phy'].data.enable_routing = False

    if address_family == "None":
        log.info("IP addressing disabled, skipping IPv4")
        anm.add_overlay("ipv4")  # create empty so rest of code follows
        g_phy.update(g_phy, use_ipv4=False)
    elif address_family in ("v4", "dual_stack"):
        build_ipv4(anm, infrastructure=True)
        g_phy.update(g_phy, use_ipv4=True)
    elif address_family == "v6":
        # Allocate v4 loopbacks for router ids
        build_ipv4(anm, infrastructure=False)
        g_phy.update(g_phy, use_ipv4=False)

    # TODO: Create collision domain overlay for ip addressing - l2 overlay?
    if address_family == "None":
        log.info("IP addressing disabled, not allocating IPv6")
        anm.add_overlay("ipv6")  # create empty so rest of code follows
        g_phy.update(g_phy, use_ipv6=False)
    elif address_family in ("v6", "dual_stack"):
        build_ipv6(anm)
        g_phy.update(g_phy, use_ipv6=True)
    else:
        anm.add_overlay("ipv6")  # placeholder for compiler logic

    default_igp = g_in.data.igp or "ospf"
    ank_utils.set_node_default(g_in, igp=default_igp)
    ank_utils.copy_attr_from(g_in, g_phy, "igp")

    ank_utils.copy_attr_from(g_in, g_phy, "include_csr")

    try:
        from autonetkit_cisco import build_network as cisco_build_network
    except ImportError, error:
        log.debug("Unable to load autonetkit_cisco %s" % error)
    else:
        cisco_build_network.pre_design(anm)

    # log.info("Building IGP")
    from autonetkit.design.igp import build_igp
    build_igp(anm)

    # log.info("Building BGP")
    from autonetkit.design.bgp import build_bgp
    build_bgp(anm)
    # autonetkit.update_vis(anm)

    from autonetkit.design.mpls import mpls_te, mpls_oam
    mpls_te(anm)
    mpls_oam(anm)

# post-processing
    if anm['phy'].data.enable_routing:
        from autonetkit.design.mpls import (mark_ebgp_vrf,
                                            build_ibgp_vpn_v4)
        mark_ebgp_vrf(anm)
        build_ibgp_vpn_v4(anm)  # build after bgp as is based on
    # autonetkit.update_vis(anm)

    try:
        from autonetkit_cisco import build_network as cisco_build_network
    except ImportError, error:
        log.debug("Unable to load autonetkit_cisco %s", error)
    else:
        cisco_build_network.post_design(anm)

    # log.info("Finished building network")
    return anm


def build(input_graph):
    """Main function to build network overlay topologies"""
    anm = None
    anm = initialise(input_graph)
    anm = apply_design_rules(anm)
    return anm

def build_phy(anm):
    """Build physical overlay"""
    g_in = anm['input']
    g_phy = anm['phy']

    g_phy.data.enable_routing = g_in.data.enable_routing
    if g_phy.data.enable_routing is None:
        g_in.data.enable_routing = True  # default if not set

    g_phy.add_nodes_from(g_in, retain=['label', 'update', 'device_type',
                                       'asn', 'specified_int_names', 'x', 'y',
                                       'device_subtype', 'platform', 'host', 'syntax'])

    if g_in.data.Creator == "Topology Zoo Toolset":
        ank_utils.copy_attr_from(g_in, g_phy, "Network")

    ank_utils.set_node_default(g_phy, Network=None)
    g_phy.add_edges_from(g_in.edges(type="physical"))
    # TODO: make this automatic if adding to the physical graph?

    ank_utils.set_node_default(g_phy, use_ipv4=False, use_ipv6=False)
    ank_utils.copy_attr_from(g_in, g_phy, "custom_config_global",
                             dst_attr="custom_config")

    for node in g_phy:
        if node['input'].custom_config_loopback_zero:
            lo_zero_config = node['input'].custom_config_loopback_zero
            node.loopback_zero.custom_config = lo_zero_config
        custom_config_phy_ints = node['input'].custom_config_phy_ints
        for interface in node:
            if custom_config_phy_ints:
                interface.custom_config = custom_config_phy_ints
            specified_id = interface['input'].get("specified_id")
            if specified_id:
                interface.specified_id = specified_id  # map across

    #TODO: tidy this code up
    for node in g_phy:
        for interface in node:
            remote_edges = interface.edges()
            if len(remote_edges):
                interface.description = 'to %s' \
                % remote_edges[0].dst.label


def build_conn(anm):
    """Build connectivity overlay"""
    g_in = anm['input']
    g_conn = anm.add_overlay("conn", directed=True)
    g_conn.add_nodes_from(g_in, retain=['label'])
    g_conn.add_edges_from(g_in.edges(type="physical"))

    return



#!/usr/bin/python
# -*- coding: utf-8 -*-
import time

import autonetkit.log as log
import networkx as nx
from autonetkit.anm.graph import NmGraph
from autonetkit.anm.ank_element import AnkElement


class NetworkModel(AnkElement):

    """"""

    def __init__(self, all_multigraph=False):
        """"""

        self.all_multigraph = all_multigraph
        self._overlays = {}
        self.add_overlay('input')
        self.add_overlay('phy')
        self.add_overlay('graphics')
        self.add_overlay('_dependencies', directed=True)

        self.label_seperator = '_'
        self.label_attrs = ['label']
        self._build_node_label()
        self.timestamp = time.strftime('%Y%m%d_%H%M%S',
                                       time.localtime())

        # TODO: make this a proper method
        self.init_logging("ANM")

    def __repr__(self):
        """"""

        return 'ANM %s' % self.timestamp

    def __len__(self):
        return len(self._overlays)

    @property
    def overlay_nx_graphs(self):
        """"""

        return self._overlays

    def has_overlay(self, overlay_id):
        """"""

        return overlay_id in self._overlays

    def dump(self):
        import autonetkit.ank_json as ank_json
        data = ank_json.jsonify_anm(self)

        # data = data.replace("\\n", "\n")
        # data = data.replace('\\"', '\"')

        return data

    def save(self):
        """"""

        # TODO: take optional filename as parameter

        import autonetkit.ank_json as ank_json
        import os
        import gzip
        archive_dir = os.path.join('versions', 'anm')
        if not os.path.isdir(archive_dir):
            os.makedirs(archive_dir)

        data = ank_json.jsonify_anm(self)
        json_file = 'anm_%s.json.gz' % self.timestamp
        json_path = os.path.join(archive_dir, json_file)
        log.debug('Saving to %s' % json_path)
        with gzip.open(json_path, 'wb') as json_fh:
            json_fh.write(data)

    def restore_latest(self, directory=None):
        """Restores latest saved ANM"""

        import os
        import glob
        if not directory:
            directory = os.path.join('versions', 'anm')

        glob_dir = os.path.join(directory, '*.json.gz')
        pickle_files = glob.glob(glob_dir)
        pickle_files = sorted(pickle_files)
        try:
            latest_file = pickle_files[-1]
        except IndexError:

            # No files loaded

            log.warning('No previous ANM saved. Please compile new ANM')
            return
        self.restore(latest_file)

    def restore(self, pickle_file):
        """"""

        import json
        import gzip
        import autonetkit.ank_json as ank_json
        log.debug('Restoring %s' % pickle_file)
        with gzip.open(pickle_file, 'r') as filehandle:
            data = json.load(filehandle)
            for (overlay_id, graph_data) in data.items():

                self._overlays[overlay_id] = \
                    ank_json.ank_json_loads(graph_data)

        ank_json.rebind_interfaces(self)

    def restore_from_json(self, in_data):
        import json
        import autonetkit.ank_json as ank_json
        data = json.loads(in_data)
        for (overlay_id, graph_data) in data.items():

            self._overlays[overlay_id] = \
                ank_json.ank_json_loads(graph_data)

            ank_json.rebind_interfaces(self)

    @property
    def _phy(self):
        """"""

        return NmGraph(self, 'phy')

    def initialise_graph(self, graph):
        """Sets input graph. Converts to undirected.
        Initialises graphics overlay."""

        # TODO: remove this dependency from workflow

        graph = nx.Graph(graph)
        g_graphics = self['graphics']
        g_in = self.add_overlay('input', graph=graph, directed=False)
        g_graphics.add_nodes_from(g_in, retain=['x', 'y', 'device_type',
                                                'device_subtype', 'pop', 'label', 'asn'])
        return g_in

    def initialise_input(self, graph):
        """Initialises input graph"""

        # remove current input
        del self._overlays['input']

        overlay = self.add_overlay('input', graph=graph)
        overlay.allocate_input_interfaces()
        return overlay

    def add_overlay(self, name, nodes=None, graph=None,
                    directed=False, multi_edge=False, retain=None):
        """Adds overlay graph of name name"""
        # TODO: refactor this logic
        log.debug("Adding overlay %s" % name)

        multi_edge = multi_edge or self.all_multigraph

        if graph:
            if not directed and graph.is_directed():
                if multi_edge or graph.is_multigraph():
                    new_graph = nx.MultiGraph(graph)
                else:
                    # TODO: put into dev log
                    log.debug('Converting graph %s to undirected' % name)
                    new_graph = nx.Graph(graph)
            else:
                if multi_edge or graph.is_multigraph():
                    new_graph = nx.MultiGraph(graph)
                else:
                    new_graph = nx.Graph(graph)
        elif directed:

            if multi_edge:
                new_graph = nx.MultiDiGraph()
            else:
                new_graph = nx.DiGraph()
        else:
            if multi_edge:
                new_graph = nx.MultiGraph()
            else:
                new_graph = nx.Graph()

        self._overlays[name] = new_graph
        overlay = NmGraph(self, name)

        if nodes:
            retain = retain or []  # default is an empty list
            overlay.add_nodes_from(nodes, retain)

        return overlay

    def __iter__(self):
        return iter(NmGraph(self, name) for name in self.overlays())

    def overlays(self):
        # TODO: rename to overlay ids
        """"""

        return self._overlays.keys()

    def devices(self, *args, **kwargs):
        """"""

        return self._phy.filter(*args, **kwargs)

    def __getitem__(self, key):
        """"""

        return NmGraph(self, key)

    def overlay(self, key):
        """"""

        return NmGraph(self, key)

    def node_label(self, node):
        """Returns node label from physical graph"""

        # TODO: refactor out node label
        return self.default_node_label(node)

    def _build_node_label(self):
        """"""

        def custom_label(node):
            """
            #TODO:  rewrite this logic/retire this function
            retval = []
            if node in self['phy']:
                for key in self.label_attrs:
                    val = self._overlays['phy'].node[node.node_id].get(key)
                    if val is not None:
                        retval.append(val)

            if len(retval):
                return self.label_seperator.join(retval)
            """


            return self.label_seperator.join(str(self._overlays['phy'
                                                                ].node[node.node_id].get(val)) for val in
                                             self.label_attrs
                                             if self._overlays['phy'].node[node.node_id].get(val) is not None)

        self.node_label = custom_label

    def set_node_label(self, seperator, label_attrs):
        """"""

        try:
            label_attrs.lower()
            label_attrs = [label_attrs]  # was a string, put into list
        except AttributeError:
            pass  # already a list

        self.label_seperator = seperator
        self.label_attrs = label_attrs



import itertools
import logging
from functools import total_ordering

import autonetkit
import autonetkit.log as log
from autonetkit.anm.interface import NmPort
from autonetkit.log import CustomAdapter
from autonetkit.anm.ank_element import AnkElement



@total_ordering
class NmNode(AnkElement):

    """NmNode"""

    def __init__(self, anm, overlay_id, node_id):

        # Set using this method to bypass __setattr__

        object.__setattr__(self, 'anm', anm)
        object.__setattr__(self, 'overlay_id', overlay_id)
# should be able to use _graph from here as anm and overlay_id are defined
        object.__setattr__(self, 'node_id', node_id)
        #logger = logging.getLogger("ANK")
        #logstring = "Node: %s" % str(self)
        #logger = CustomAdapter(logger, {'item': logstring})
        logger = log
        object.__setattr__(self, 'log', logger)
        self.init_logging("node")

    def __hash__(self):
        """"""

        return hash(self.node_id)

    def __nonzero__(self):
        """Allows for checking if node exists

        >>> anm = autonetkit.topos.house()
        >>> r1 = anm['phy'].node("r1")
        >>> bool(r1)
        True
        >>> r_test = anm['phy'].node("test")
        >>> bool(r_test)
        False

        """

        return self.node_id in self._graph

    def __iter__(self):
        """Shortcut to iterate over the physical interfaces of this node

        >>> anm = autonetkit.topos.house()
        >>> r1 = anm['phy'].node("r1")
        >>> list(r1)
        [eth0.r1, eth1.r1]


        """

        return iter(self.interfaces(category='physical'))

    def __len__(self):
        """Number of phyiscal interfaces of a node

        >>> anm = autonetkit.topos.house()
        >>> r1 = anm['phy'].node("r1")
        >>> len(r1)
        2
        >>> r2 = anm['phy'].node("r2")
        >>> len(r2)
        3


        """
        return len(list(self.__iter__()))

    def __eq__(self, other):
        """

        >>> anm = autonetkit.topos.house()
        >>> r1 = anm['phy'].node("r1")
        >>> rb = anm['phy'].node("r1")
        >>> r1 == rb
        True
        >>> r2 = anm['phy'].node("r2")
        >>> r1 == r2
        False

        Can also compare to a label

        >>> r1 == "r1"
        True
        >>> r1 == "r2"
        False

        """

        try:
            return self.node_id == other.node_id
        except AttributeError:
            return self.node_id == other  # eg compare Node to label

    def __ne__(self, other):
        """

        >>> anm = autonetkit.topos.house()
        >>> r1 = anm['phy'].node("r1")
        >>> r2 = anm['phy'].node("r2")
        >>> r1 != r2
        True


        """
        return not self.__eq__(other)

    @property
    def loopback_zero(self):
        """

        >>> anm = autonetkit.topos.house()
        >>> r1 = anm['phy'].node("r1")
        >>> lo_zero = r1.loopback_zero

        """

        # TODO: initialise id for loopback zero?
        # TODO: Set category for loopback zero to be "loopback"

        return (i for i in self.interfaces('is_loopback_zero')).next()

    def physical_interfaces(self, *args, **kwargs):
        """"""
        kwargs['category'] = "physical"
        return self.interfaces(*args, **kwargs)

    def loopback_interfaces(self, *args, **kwargs):
        #TODO: allow abiility to skip loopback zero
        """"""
        kwargs['category'] = "loopback"
        return self.interfaces(*args, **kwargs)

    def is_multigraph(self):
        return self._graph.is_multigraph()

    def __lt__(self, other):
        """

        >>> anm = autonetkit.topos.house()
        >>> r1 = anm['phy'].node("r1")
        >>> r2 = anm['phy'].node("r2")
        >>> r1 < r2
        True

        """

# want [r1, r2, ..., r11, r12, ..., r21, r22] not [r1, r11, r12, r2, r21, r22]
# so need to look at numeric part
        import string
        # TODO: use human sort from StackOverflow

        # sort on label if available
        if self.label is not None:
            self_node_id = self.label
        else:
            self_node_id = self.node_id

        if other.label is not None:
            other_node_id = other.label
        else:
            other_node_id = other_node_id

        try:
            self_node_string = [x for x in self_node_id if x
                                not in string.digits]
            other_node_string = [x for x in self_node_id if x
                                 not in string.digits]
        except TypeError:

            # e.g. non-iterable category, such as an int node_id

            pass
        else:
            if self_node_string == other_node_string:
                self_node_id = """""".join([x for x in self_node_id
                                            if x in string.digits])
                other_node_id = """""".join([x for x in other_node_id
                                             if x in string.digits])
                try:
                    self_node_id = int(self_node_id)
                except ValueError:
                    pass  # not a number
                try:
                    other_node_id = int(other_node_id)
                except ValueError:
                    pass  # not a number

        return (self.asn, self_node_id) < (other.asn, other_node_id)

    def _next_int_id(self):
        """"""

# returns next free interface I

        for int_id in itertools.count(1):  # start at 1 as 0 is loopback
            if int_id not in self._ports:
                return int_id

    # TODO: interface function access needs to be cleaned up

    def _add_interface(self, description=None, category='physical',
                       **kwargs):
        """"""

        data = dict(kwargs)

        if self.overlay_id != 'phy' and self['phy']:
            next_id = self['phy']._next_int_id()
            self['phy']._ports[next_id] = {'category': category,
                                           'description': description}

            # TODO: fix this workaround for not returning description from phy
            # graph

            data['description'] = description
        else:
            next_id = self._next_int_id()
            data['category'] = category  # store category on node
            data['description'] = description

        self._ports[next_id] = data
        return next_id

    def _sync_loopbacks(self, interface_id):
        """Syncs a newly added loopback across all overlays the
        node is in"""
        for overlay in self.anm:
            if self not in overlay:
                continue

            if overlay._overlay_id in ("phy", self.overlay_id):
                # don't copy to self or phy as already there
                continue

            if overlay._overlay_id in ("graphics"):
                """
                TODO: debug why get problem for graphics
                 File "/autonetkit/autonetkit/anm/node.py", line 257, in _sync_loopbacks
                    o_node._ports[interface_id] = {"category": "loopback"}
                IndexError: list assignment index out of range
                None
                """
                # don't copy to self or phy as already there
                continue

            o_node = overlay.node(self)
            if interface_id in o_node._ports:
                # something has gone wrong - allocated the id over the top
                # shouldn't happen since the next free id is from the phy

                log.warning("Internal consistency error with copying loopback "
                    "interface %s to %s in %s", interface_id, self, overlay)
                continue

            o_node._ports[interface_id] = {"category": "loopback"}

    def add_loopback(self, *args, **kwargs):
        '''Public function to add a loopback interface'''

        interface_id = self._add_interface(category='loopback', *args,
                                           **kwargs)

        #TODO: want to add loopbacks to all overlays the node is in
        self._sync_loopbacks(interface_id)
        return NmPort(self.anm, self.overlay_id,
                      self.node_id, interface_id)

    def add_interface(self, *args, **kwargs):
        """Public function to add interface"""

        interface_id = self._add_interface(*args, **kwargs)
        return NmPort(self.anm, self.overlay_id,
                      self.node_id, interface_id)

    def interfaces(self, *args, **kwargs):
        """Public function to view interfaces"""

        def filter_func(interface):
            """Filter based on args and kwargs"""

            return all(getattr(interface, key) for key in args) \
                and all(getattr(interface, key) == val for (key,
                                                            val) in kwargs.items())

        all_interfaces = iter(NmPort(self.anm,
                                     self.overlay_id, self.node_id,
                                     interface_id) for interface_id in
                              self._interface_ids())

        retval = [i for i in all_interfaces if filter_func(i)]
        return retval

    def interface(self, key):
        """Returns interface based on interface id"""

        try:
            if key.interface_id in self._interface_ids():
                return NmPort(self.anm, self.overlay_id,
                              self.node_id, key.interface_id)
        except AttributeError:

            # try with key as id

            try:
                if key in self._interface_ids():
                    return NmPort(self.anm, self.overlay_id,
                                  self.node_id, key)
            except AttributeError:

                # no match for either

                log.warning('Unable to find interface %s in %s '
                            % (key, self))
                return None

        # try searching for the "id" attribute of the interface eg
        # GigabitEthernet0/0 if set
        search = list(self.interfaces(id=key))
        # TODO: warn if more than one match ie len > 1
        if len(search):
            return search[0]  # first result

        search = list(self.interfaces(description=key))
        # TODO: warn if more than one match ie len > 1
        if len(search):
            return search[0]  # first result

    def _interface_ids(self):
        """Returns interface ids for this node"""
        # TODO: use from this layer, otherwise can get errors iterating when eg
        # vrfs
        return self._ports.keys()

        if self.overlay_id != 'phy' and self['phy']:

            # graph isn't physical, and node exists in physical graph -> use
            # the interface mappings from phy

            return self['phy']._graph.node[self.node_id]['_ports'
                                                         ].keys()
        else:
            try:
                return self._ports.keys()
            except KeyError:
                self.log.debug('No interfaces initialised')
                return []

    @property
    def _ports(self):
        """Returns underlying interface dict"""

        try:
            return self._graph.node[self.node_id]['_ports']
        except KeyError:
            self.log.debug('No interfaces initialised for')
            return []

    @property
    def _graph(self):
        """Return graph the node belongs to"""

        try:
            return self.anm.overlay_nx_graphs[self.overlay_id]
        except Exception, e:
            log.warning("Error accessing overlay %s for node %s: %s" %
                        (self.overlay_id, self.node_id, e))

    @property
    def _nx_node_data(self):
        """Return NetworkX node data for the node"""
        try:
            return self._graph.node[self.node_id]
        except Exception, e:
            log.warning("Error accessing node data %s for node %s: %s" %
                        (self.overlay_id, self.node_id, e))

    def is_router(self):
        """Either from this graph or the physical graph

        >>> anm = autonetkit.topos.mixed()
        >>> r1 = anm['phy'].node("r1")
        >>> r1.is_router()
        True
        >>> s1 = anm['phy'].node("s1")
        >>> s1.is_router()
        False
        >>> sw1 = anm['phy'].node("sw1")
        >>> sw1.is_router()
        False

        """
        #self.log_info("add int")

        return self.device_type == 'router' or self['phy'].device_type \
            == 'router'

    def is_hub(self):
        """Either from this graph or the physical graph

        """
        #self.log_info("add int")

        return self.device_type == 'hub' or self['phy'].device_type \
            == 'hub'

    def is_device_type(self, device_type):
        """Generic user-defined cross-overlay search for device_type
        either from this graph or the physical graph"""

        return self.device_type == device_type or self['phy'].device_type \
            == device_type

    def is_switch(self):
        """Returns if device is a switch

        >>> anm = autonetkit.topos.mixed()
        >>> r1 = anm['phy'].node("r1")
        >>> r1.is_switch()
        False
        >>> s1 = anm['phy'].node("s1")
        >>> s1.is_switch()
        False
        >>> sw1 = anm['phy'].node("sw1")
        >>> sw1.is_switch()
        True

        """

        return self.device_type == 'switch' or self['phy'].device_type \
            == 'switch'

    def is_server(self):
        """Returns if device is a server

        >>> anm = autonetkit.topos.mixed()
        >>> r1 = anm['phy'].node("r1")
        >>> r1.is_server()
        False
        >>> s1 = anm['phy'].node("s1")
        >>> s1.is_server()
        True
        >>> sw1 = anm['phy'].node("sw1")
        >>> sw1.is_server()
        False

        """

        return self.device_type == 'server' or self['phy'].device_type \
            == 'server'

    def is_l3device(self):
        """Layer 3 devices: router, server, cloud, host
        ie not switch

        >>> anm = autonetkit.topos.mixed()
        >>> r1 = anm['phy'].node("r1")
        >>> r1.is_l3device()
        True
        >>> s1 = anm['phy'].node("s1")
        >>> s1.is_l3device()
        True
        >>> sw1 = anm['phy'].node("sw1")
        >>> sw1.is_l3device()
        False

        """
        return self.is_router() or self.is_server()

    def __getitem__(self, key):
        """Get item key"""

        return NmNode(self.anm, key, self.node_id)

    @property
    def raw_interfaces(self):
        """Direct access to the interfaces dictionary, used by ANK modules"""
        return self._ports

    @raw_interfaces.setter
    def raw_interfaces(self, value):
        self._ports = value

    @property
    def asn(self):
        """Returns ASN of this node

        >>> anm = autonetkit.topos.house()
        >>> r1 = anm['phy'].node("r1")
        >>> r1.asn
        1
        >>> r5 = anm['phy'].node("r5")
        >>> r5.asn
        2

        """
        # TODO: make a function (not property)
        # TODO: refactor, for nodes created such as virtual switches

        try:
            return self._graph.node[self.node_id]['asn']  # not in this graph
        except KeyError:

            # try from phy

            try:
                return self.anm.overlay_nx_graphs['phy'
                                                  ].node[self.node_id]['asn']
            except KeyError:
                if self.node_id not in self.anm.overlay_nx_graphs['phy'
                                                                  ]:
                    message = \
                        'Node id %s not found in physical overlay' \
                        % self.node_id
                    if self.overlay_id == 'input':

                        # don't warn, most likely node not copied across

                        log.debug(message)
                    else:
                        log.debug(message)
                    return

    @asn.setter
    def asn(self, value):
        # TODO: make a function (not property)

        # TODO: double check this logic

        try:
            self.anm.overlay_nx_graphs['phy'].node[self.node_id]['asn'
                                                                 ] = value
        except KeyError:

            # set ASN directly on the node, eg for collision domains

            self._graph.node[self.node_id]['asn'] = value

    @property
    def id(self):
        """Returns node id"""

        return self.node_id

    @property
    def _overlay(self):
        """Access overlay graph for this node"""

        from autonetkit.anm.graph import NmGraph
        return NmGraph(self.anm, self.overlay_id)

    def degree(self):
        """Returns degree of node

        >>> anm = autonetkit.topos.house()
        >>> r1 = anm['phy'].node("r1")
        >>> r1.degree()
        2
        >>> r2 = anm['phy'].node("r2")
        >>> r2.degree()
        3

        """

        # TODO: add example for multi-edge

        return self._graph.degree(self.node_id)

    def neighbors(self, *args, **kwargs):
        """Returns neighbors of node

        >>> anm = autonetkit.topos.house()
        >>> r1 = anm['phy'].node("r1")
        >>> r1.neighbors()
        [r2, r3]
        >>> r2 = anm['phy'].node("r2")
        >>> r2.neighbors()
        [r4, r1, r3]

        Can also filter

        >>> r2.neighbors(asn=1)
        [r1, r3]
        >>> r2.neighbors(asn=2)
        [r4]


        """

        neighs = list(NmNode(self.anm, self.overlay_id, node)
                      for node in self._graph.neighbors(self.node_id))

        return self._overlay.filter(neighs, *args, **kwargs)

    def neighbor_interfaces(self, *args, **kwargs):
        """

        >>> anm = autonetkit.topos.house()
        >>> r1 = anm['phy'].node("r1")
        >>> r1.neighbor_interfaces()
        [eth0.r2, eth0.r3]
        >>> r2 = anm['phy'].node("r2")
        >>> r2.neighbor_interfaces()
        [eth0.r4, eth0.r1, eth1.r3]

        """

        # TODO: implement filtering for args and kwargs

        if len(args) or len(kwargs):
            log.warning("Attribute-based filtering not currently",
                        "supported for neighbor_interfaces")

        return list(edge.dst_int for edge in self.edges())

    @property
    # TODO: make a function to reflect dynamic nature: constructed from other
    # attributes
    def label(self):
        """Returns node label (mapped from ANM)"""

        return self.__repr__()

    def dump(self):
        """Dump attributes of this node"""

        data = dict(self._nx_node_data)
        try:
            del data['_ports']
        except KeyError:
            pass  # no interfaces set
        return str(data)

    def edges(self, *args, **kwargs):
        """Edges to/from this node

        >>> anm = autonetkit.topos.house()
        >>> r1 = anm['phy'].node("r1")
        >>> r1.edges()
        [(r1, r2), (r1, r3)]
        >>> r2 = anm['phy'].node("r2")
        >>> r2.edges()
        [(r2, r4), (r2, r1), (r2, r3)]

        """

        return list(self._overlay.edges(self, *args, **kwargs))

    def __str__(self):
        return str(self.__repr__())

    def __repr__(self):
        """Try label if set in overlay, otherwise from physical,
        otherwise node id"""

        try:
            return self.anm.node_label(self)
        except KeyError:
            try:
                return self._graph.node[self.node_id]['label']
            except KeyError:
                return self.node_id  # node not in physical graph
        except AttributeError:
            return self.node_id

    def __getattr__(self, key):
        """Returns node property
        This is useful for accesing attributes passed through from graphml"""
        # TODO: refactor/document this logic

        try:
            node_data = self._graph.node[self.node_id]
        except KeyError:
            # TODO: only carry out this logic if "Strict mode"
            if key == "device_type":
                # TODO: tidy accessors so this doesn't occur, and remove the
                # suppress
                return
            self.log.debug("Cannot access %s: node not present in %s"
                           % (key, self.overlay_id))
            return
        else:
            try:
                result = node_data[key]
                return result
            except KeyError:
                #TODO: check if in self['phy'[ here]
                if self.overlay_id != "phy":
                    if key == "device_type":
                        return self['phy'].device_type
                    if key == "device_subtype":
                        return self['phy'].device_subtype

                # from http://stackoverflow.com/q/2654113
                #self.log.debug(
                    #"Accessing unset attribute %s in %s" % (key,
                                                            #self.overlay_id))
                return

        # map through to phy

    def get(self, key):
        """For consistency, node.get(key) is neater than getattr(node, key)"""

        return getattr(self, key)

    def __setattr__(self, key, val):
        """Sets node property
        This is useful for accesing attributes passed through from graphml"""

        """TODO:
        TODO: look at mapping the object __dict__ straight to the graph.node[self.node_id]
        TODO: fix wrt using @x.setter won't work due to following:
        as per
        http://docs.python.org/2/reference/datamodel.html#customizing-attribute-access
        """

        # TODO: fix workaround for asn

        if key == 'asn':
            object.__setattr__(self, 'asn', val)

        if key == 'raw_interfaces':
            object.__setattr__(self, 'raw_interfaces', val)

        try:
            self._graph.node[self.node_id][key] = val
        except KeyError:
            self._graph.add_node(self.node_id)
            self.set(key, val)

    def set(self, key, val):
        """For consistency, node.set(key, value) is neater
        than setattr(node, key, value)"""

        return self.__setattr__(key, val)



import autonetkit.ank as ank_utils
import autonetkit.log as log

"""TODO: map the log info/warning/debug to functions here
# which then map to the appropriate log function, so that can handle either be verbose or not verbose to console with test information.
"""


def validate(anm):
    log.debug("Validating overlay topologies")
    tests_passed = True
    tests_passed = validate_ipv4(anm) and tests_passed

    try:
        from autonetkit_cisco import ank_validate as cisco_validate
    except ImportError, e:
        log.debug("Unable to load autonetkit_cisco %s" % e)
    else:
        cisco_validate.validate(anm)

    validate_ibgp(anm)
    validate_igp(anm)
    check_for_selfloops(anm)
    all_nodes_have_asn(anm)

    if tests_passed:
        log.debug("All validation tests passed.")
    else:
        log.warning("Some validation tests failed.")


def check_for_selfloops(anm):
    # checks each overlay for selfloops
    for overlay in anm:
        selfloop_count = overlay._graph.number_of_selfloops()
        if selfloop_count > 0:
            log.warning("%s has %s self-loops" % (overlay, selfloop_count))


def all_nodes_have_asn(anm):
    g_phy = anm['phy']
    for node in g_phy.l3devices():
        if node.asn is None:
            log.warning("No ASN set for physical device %s" % node)


def validate_ibgp(anm):
    import networkx as nx
    # TODO: repeat for ibgp v6
    # TODO: test if overlay is present, if not then warn
    if not anm.has_overlay("ibgp_v4"):
        return  # no ibgp v4  - eg if ip addressing disabled

    g_ibgp_v4 = anm['ibgp_v4']

    for asn, devices in ank_utils.groupby("asn", g_ibgp_v4):
        asn_subgraph = g_ibgp_v4.subgraph(devices)
        graph = asn_subgraph._graph
        # get subgraph
        if not nx.is_strongly_connected(graph):
            g_ibgp_v4.log.warning(
                "iBGP v4 topology for ASN%s is disconnected" % asn)
            # TODO: list connected components - but not the primary?
        else:
            g_ibgp_v4.log.debug(
                "iBGP v4 topology for ASN%s is connected" % asn)


def validate_igp(anm):
    import networkx as nx
    # TODO: test if overlay is present, if not then warn
    if not anm.has_overlay("igp"):
        return  # no ibgp v4  - eg if ip addressing disabled

    g_igp = anm['igp']

    for asn, devices in ank_utils.groupby("asn", g_igp):
        if asn is None:
            continue
        asn_subgraph = g_igp.subgraph(devices)
        graph = asn_subgraph._graph
        # get subgraph
        if not nx.is_connected(graph):
            g_igp.log.warning("IGP topology for ASN%s is disconnected" % asn)
            # TODO: list connected components - but not the primary?
        else:
            g_igp.log.debug("IGP topology for ASN%s is connected" % asn)


def all_same(items):
    # based on http://stackoverflow.com/q/3787908
    # use all with generator as shortcuts to false as soon as one invalid
    # TODO: place this into generic ANK functions, use similar shortcut for
    # speed elsewhere, in other utility functions
    return all(x == items[0] for x in items)


def all_unique(items):
    # based on http://stackoverflow.com/q/3787908
    # shortcuts for efficiency
    seen = set()
    return not any(i in seen or seen.add(i) for i in items)


def duplicate_items(items):
    unique = set(items)
    counts = {i: items.count(i) for i in unique}
    return [i for i in counts if counts[i] > 1]

# TODO: add high-level symmetry, anti-summetry, uniqueness, etc functions
# as per NCGuard

# TODO: make generic interface equal or unique function that takes attr


def validate_ipv4(anm):
    # TODO: make this generic to also handle IPv6
    if not anm.has_overlay("ipv4"):
        log.debug("No IPv4 overlay created, skipping ipv4 validation")
        return
    g_ipv4 = anm['ipv4']
    # interface IP uniqueness
    tests_passed = True

    # TODO: only include bound interfaces

    # check globally unique ip addresses
    all_ints = [i for n in g_ipv4.l3devices()
                for i in n.physical_interfaces()
                if i.is_bound]  # don't include unbound interfaces
    all_int_ips = [i.ip_address for i in all_ints if i.ip_address]

    if all_unique(all_int_ips):
        g_ipv4.log.debug("All interface IPs globally unique")
    else:
        tests_passed = False
        duplicates = duplicate_items(all_int_ips)
        duplicate_ips = set(duplicate_items(all_int_ips))
        duplicate_ints = [n for n in all_ints
                          if n.ip_address in duplicate_ips]
        duplicates = ", ".join("%s: %s" % (i.node, i.ip_address)
                               for i in duplicate_ints)
        g_ipv4.log.warning("Global duplicate IP addresses %s" % duplicates)

    for bc in g_ipv4.nodes("broadcast_domain"):
        bc.log.debug("Verifying subnet and interface IPs")
        if not bc.allocate:
            log.debug("Skipping validation of manually allocated broadcast "
                "domain %s" % bc)
            continue

        neigh_ints = list(bc.neighbor_interfaces())
        neigh_ints = [i for i in neigh_ints if i.node.is_l3device()]
        neigh_int_subnets = [i.subnet for i in neigh_ints]
        if all_same(neigh_int_subnets):
            # log ok
            pass
        else:
            subnets = ", ".join("%s: %s" % (i.node, i.subnets)
                                for i in neigh_int_subnets)
            tests_passed = False
            log.warning("Different subnets on %s. %s" %
                        (bc, subnets))
            # log warning

        ip_subnet_mismatches = [i for i in neigh_ints
                                if i.ip_address not in i.subnet]
        if len(ip_subnet_mismatches):
            tests_passed = False
            mismatches = ", ".join("%s not in %s on %s" %
                                   (i.ip_address, i.subnet, i.node)
                                   for i in ip_subnet_mismatches)
            bc.log.warning("Mismatched IP subnets: %s" %
                           mismatches)
        else:
            bc.log.debug("All subnets match")

        neigh_int_ips = [i.ip_address for i in neigh_ints]
        if all_unique(neigh_int_ips):
            bc.log.debug("All interface IP addresses are unique")
            duplicates = duplicate_items(neigh_int_ips)
        else:
            tests_passed = False
            duplicate_ips = set(duplicate_items(neigh_int_ips))
            duplicate_ints = [n for n in neigh_ints
                              if n.ip_address in duplicate_ips]
            duplicates = ", ".join("%s: %s" % (i.node, i.ip_address)
                                   for i in duplicate_ints)
            bc.log.warning("Duplicate IP addresses: %s" % duplicates)

    if tests_passed:
        g_ipv4.log.debug("All IP tests passed.")
    else:
        g_ipv4.log.warning("Some IP tests failed.")

    return tests_passed



import sys


def convert():
    filename = sys.argv[1]
    with open(filename, "r") as fh:
        data = fh.read()

    from autonetkit.build_network import build, load
    graph = load(data)
    anm = build(graph)


    import autonetkit
    g_in_nx = anm['input']._graph

    output = autonetkit.load.load_json.nx_to_simple(g_in_nx)

if __name__ == "__main__":
    main()


import logging

import autonetkit.log as log
from autonetkit.log import CustomAdapter
from autonetkit.anm.ank_element import AnkElement


class NmPort(AnkElement):

    def __init__(self, anm, overlay_id, node_id, interface_id):
        object.__setattr__(self, 'anm', anm)
        object.__setattr__(self, 'overlay_id', overlay_id)
        object.__setattr__(self, 'node_id', node_id)
        object.__setattr__(self, 'interface_id', interface_id)
        #logger = logging.getLogger("ANK")
        #logstring = "Interface: %s" % str(self)
        #logger = CustomAdapter(logger, {'item': logstring})
        logger = log
        object.__setattr__(self, 'log', logger)
        self.init_logging("port")


    def __key(self):
        """Note: key doesn't include overlay_id to allow fast cross-layer comparisons"""

        # based on http://stackoverflow.com/q/2909106

        return (self.interface_id, self.node_id)

    def __hash__(self):
        """"""

        return hash(self.__key())


    def __deepcopy__(self, memo):
        #TODO: workaround - need to fix
       # from http://stackoverflow.com/questions/1500718/what-is-the-right-way-to-override-the-copy-deepcopy-operations-on-an-object-in-p
       pass


    def __repr__(self):
        description = self.id or self.description
        return '%s.%s' % (description, self.node)

    def __eq__(self, other):
        return self.__key() == other.__key()

    def __nonzero__(self):

        # TODO: work out why description and category being set/copied to each
        # overlay

        try:
            interface = self._interface
        except KeyError:
            return False

        return len(interface) > 0  # if interface data set

    def __lt__(self, other):

        # TODO: check how is comparing the node

        return (self.node, self.interface_id) < (other.node,
                                                 other.interface_id)


    @property
    def id(self):
        """Returns  id of node, falls-through to phy if not set on this overlay

        """
        #TODO: make generic function for fall-through properties stored on the anm
        key = "id"
        if key in self._interface:
                return self._interface.get(key)
        else:
            if self.overlay_id == "input":
                return  # Don't fall upwards from input -> phy as may not exist
            if self.overlay_id == "phy":
                return  # Can't fall from phy -> phy (loop)

            # try from phy

            if not self.node['phy']:
                # node not in phy
                return

            try:
                #return self.anm.overlay_nx_graphs['phy'].node[self.node_id]['asn']
                return self['phy'].id
            except KeyError:
                return # can't get from this overlay or phy -> not found

    @property
    def is_bound(self):
        # TODO: make this a function
        """Returns if this interface is bound to an edge on this layer"""

        return len(self.edges()) > 0

    def __str__(self):
        return self.__repr__()

    @property
    def _graph(self):
        """Return graph the node belongs to"""

        return self.anm.overlay_nx_graphs[self.overlay_id]

    @property
    def _node(self):
        """Return graph data the node belongs to"""

        return self._graph.node[self.node_id]

    @property
    def _interface(self):
        """Return data dict for the interface"""
        try:
            return self.node.raw_interfaces[self.interface_id]
        except KeyError:
            if not self.node_id in self._graph:
                # node not in overlay
                return
                log.warning("Unable to access interface %s in %s",
                            "node %s not present in overlay" % (self.interface_id,
                                                                self.overlay_id, self.node_id))
                return

            log.warning('Unable to find interface %s in %s'
                        % (self.interface_id, self.node))
            return None

    @property
    def phy(self):
        if self.overlay_id == 'phy':
            return self
        return NmPort(self.anm, 'phy', self.node_id, self.interface_id)

    def __getitem__(self, overlay_id):
        """Returns corresponding interface in specified overlay"""

        if not self.anm.has_overlay(overlay_id):
            log.warning('Trying to access interface %s for non-existent overlay %s'
                        , self, overlay_id)
            return None

        if not self.node_id in self.anm.overlay_nx_graphs[overlay_id]:
            log.debug('Trying to access interface %s for non-existent node %s in overlay %s'
                      , self, self.node_id, self.overlay_id)
            return None

        try:
            return NmPort(self.anm, overlay_id,
                          self.node_id, self.interface_id)
        except KeyError:
            return

    @property
    def is_loopback(self):
        """"""

        return self.category == 'loopback' or self.phy.category == 'loopback'

    @property
    def is_physical(self):
        """"""

        return self.category == 'physical' or self.phy.category == 'physical'

    @property
    def description(self):
        """"""

        return_val = None
        try:
            return_val = self._interface.get('description')
        except IndexError:
            return_val = self.interface_id

        if return_val is not None:
            return return_val

        if self.overlay_id != 'phy':  # prevent recursion
            self.phy._interface.get('description')

    @property
    def is_loopback_zero(self):
        return self.interface_id == 0 and self.is_loopback

    @property
    def category(self):
        """"""

# TODO: make 0 correctly access interface 0 -> copying problem
# TODO: this needs a bugfix rather than the below hard-coded workaround

        if self.interface_id == 0:
            return 'loopback'

        if self.overlay_id == 'input':
            return object.__getattr__(self, 'category')
        elif self.overlay_id != 'phy':

                                        # prevent recursion

            return self.phy._interface.get('category')

        retval = self._interface.get('category')
        if retval:
            return retval

    @property
    def node(self):
        """Returns parent node of this interface"""

        from autonetkit.anm.node import NmNode
        return NmNode(self.anm, self.overlay_id, self.node_id)

    def dump(self):
        return str(self._interface.items())

    def __getattr__(self, key):
        """Returns interface property"""

        try:
            return self._interface.get(key)
        except KeyError:
            return

    def get(self, key):
        """For consistency, node.get(key) is neater
        than getattr(interface, key)"""

        return getattr(self, key)

    def __setattr__(self, key, val):
        """Sets interface property"""

        try:
            self._interface[key] = val
        except KeyError, e:
            log.warning(e)

            # self.set(key, val)

    def set(self, key, val):
        """For consistency, node.set(key, value) is neater
        than setattr(interface, key, value)"""

        return self.__setattr__(key, val)

    def edges(self):
        """Returns all edges from node that have this interface ID
        This is the convention for binding an edge to an interface"""

        # edges have _interfaces stored as a dict of {node_id: interface_id, }

        valid_edges = [e for e in self.node.edges() if self.node_id
                       in e.raw_interfaces and e.raw_interfaces[self.node_id]
                       == self.interface_id]
        return list(valid_edges)

    def neighbors(self):
        """Returns interfaces on nodes that are linked to this interface
        Can get nodes using [i.node for i in interface.neighbors()]
        """

        edges = self.edges()
        return [e.dst_int for e in edges]



import autonetkit.ank as ank_utils
import autonetkit.log as log


def build_layer2(anm):
    build_layer2_base(anm)
    build_vlans(anm)
    build_layer2_conn(anm)

    check_layer2(anm)
    build_layer2_broadcast(anm)


def build_layer2_base(anm):
    g_l2 = anm.add_overlay('layer2')
    g_l1 = anm['layer1']

    g_l2.add_nodes_from(g_l1)
    g_l2.add_edges_from(g_l1.edges())
    # Don't aggregate managed switches
    for node in g_l2:
        if node['layer1'].collision_domain == True:
            node.broadcast_domain = True
            node.device_type = "broadcast_domain"

    try:
        from autonetkit_cisco import build_network as cisco_build_network
    except ImportError:
        pass
    else:
        cisco_build_network.post_layer2(anm)

    # Note: layer 2 base is much simpler since most is inherited from layer 1
    # cds


def build_layer2_conn(anm):
    g_l2 = anm['layer2']
    g_l2_conn = anm.add_overlay('layer2_conn')
    g_l2_conn.add_nodes_from(g_l2, retain="broadcast_domain")
    g_l2_conn.add_edges_from(g_l2.edges())

    broadcast_domains = g_l2_conn.nodes(broadcast_domain=True)
    exploded_edges = ank_utils.explode_nodes(g_l2_conn, broadcast_domains)

    # explode each seperately?
    for edge in exploded_edges:
        edge.multipoint = True
        edge.src_int.multipoint = True
        edge.dst_int.multipoint = True


def check_layer2(anm):
    """Sanity checks on topology"""
    from collections import defaultdict
    g_l2 = anm['layer2']

    # check for igp and ebgp on same switch
    for switch in sorted(g_l2.switches()):
        neigh_asns = defaultdict(int)
        for neigh in switch.neighbors():
            if neigh.asn is None:
                continue  # don't add if not set
            neigh_asns[neigh.asn] += 1

        # IGP if two or more neighbors share the same ASN
        is_igp = any(asns > 1 for asns in neigh_asns.values())
        # eBGP if more than one unique neigh ASN
        is_ebgp = len(neigh_asns.keys()) > 1
        if is_igp and is_ebgp:
            log.warning("Switch %s contains both IGP and eBGP neighbors",
                        switch)

    # check for multiple links from nodes to switch
    for switch in sorted(g_l2.switches()):
        for neighbor in sorted(switch.neighbors()):
            edges = g_l2.edges(switch, neighbor)
            if len(edges) > 1:
                # more than one edge between the (src, dst) pair -> parallel
                log.warning("Multiple edges (%s) between %s and device %s",
                            len(edges), switch, neighbor)


def build_layer2_broadcast(anm):
    g_l2 = anm['layer2']
    g_phy = anm['phy']
    g_graphics = anm['graphics']
    g_l2_bc = anm.add_overlay('layer2_bc')
    g_l2_bc.add_nodes_from(g_l2.l3devices())
    g_l2_bc.add_nodes_from(g_l2.switches())
    g_l2_bc.add_edges_from(g_l2.edges())

    # remove external connectors

    edges_to_split = [edge for edge in g_l2_bc.edges()
                      if edge.src.is_l3device() and edge.dst.is_l3device()]
    # TODO: debug the edges to split
    for edge in edges_to_split:
        edge.split = True  # mark as split for use in building nidb

    split_created_nodes = list(ank_utils.split(g_l2_bc, edges_to_split,
                                               retain=['split'],
                                               id_prepend='cd_'))

    # TODO: if parallel nodes, offset
    # TODO: remove graphics, assign directly
    if len(g_graphics):
        co_ords_overlay = g_graphics  # source from graphics overlay
    else:
        co_ords_overlay = g_phy  # source from phy overlay

    for node in split_created_nodes:
        node['graphics'].x = ank_utils.neigh_average(g_l2_bc, node, 'x',
                                                     co_ords_overlay) + 0.1

        # temporary fix for gh-90

        node['graphics'].y = ank_utils.neigh_average(g_l2_bc, node, 'y',
                                                     co_ords_overlay) + 0.1

        # temporary fix for gh-90

        asn = ank_utils.neigh_most_frequent(
            g_l2_bc, node, 'asn', g_phy)  # arbitrary choice
        node['graphics'].asn = asn
        node.asn = asn  # need to use asn in IP overlay for aggregating subnets

    # also allocate an ASN for virtual switches
    vswitches = [n for n in g_l2_bc.nodes()
                 if n['layer2'].device_type == "switch"
                 and n['layer2'].device_subtype == "virtual"]
    for node in vswitches:
        # TODO: refactor neigh_most_frequent to allow fallthrough attributes
        # asn = ank_utils.neigh_most_frequent(g_l2_bc, node, 'asn', g_l2)  #
        # arbitrary choice
        asns = [n['layer2'].asn for n in node.neighbors()]
        asns = [x for x in asns if x is not None]
        asn = ank_utils.most_frequent(asns)
        node.asn = asn  # need to use asn in IP overlay for aggregating subnets
        # also mark as broadcast domain

    from collections import defaultdict
    coincident_nodes = defaultdict(list)
    for node in split_created_nodes:
        coincident_nodes[(node['graphics'].x, node['graphics'].y)].append(node)

    coincident_nodes = {k: v for k, v in coincident_nodes.items()
                        if len(v) > 1}  # trim out single node co-ordinates
    import math
    for _, val in coincident_nodes.items():
        for index, item in enumerate(val):
            index = index + 1
            x_offset = 25 * math.floor(index / 2) * math.pow(-1, index)
            y_offset = -1 * 25 * math.floor(index / 2) * math.pow(-1, index)
            item['graphics'].x = item['graphics'].x + x_offset
            item['graphics'].y = item['graphics'].y + y_offset

    switch_nodes = g_l2_bc.switches()  # regenerate due to aggregated
    g_l2_bc.update(switch_nodes, broadcast_domain=True)

    # switches are part of collision domain
    g_l2_bc.update(split_created_nodes, broadcast_domain=True)

    # Assign collision domain to a host if all neighbours from same host

    for node in split_created_nodes:
        if ank_utils.neigh_equal(g_l2_bc, node, 'host', g_phy):
            node.host = ank_utils.neigh_attr(g_l2_bc, node, 'host',
                                             g_phy).next()  # first attribute

    # set collision domain IPs
    # TODO; work out why this throws a json exception
    #autonetkit.ank.set_node_default(g_l2_bc,  broadcast_domain=False)

    for node in g_l2_bc.nodes('broadcast_domain'):
        graphics_node = g_graphics.node(node)
        #graphics_node.device_type = 'broadcast_domain'
        if node.is_switch():
            # TODO: check not virtual
            node['phy'].broadcast_domain = True
        if not node.is_switch():
            # use node sorting, as accomodates for numeric/string names
            graphics_node.device_type = 'broadcast_domain'
            neighbors = sorted(neigh for neigh in node.neighbors())
            label = '_'.join(neigh.label for neigh in neighbors)
            cd_label = 'cd_%s' % label  # switches keep their names
            node.label = cd_label
            graphics_node.label = cd_label
            node.device_type = "broadcast_domain"
            node.label = node.id
            graphics_node.label = node.id

    for node in vswitches:
        node.broadcast_domain = True


def set_default_vlans(anm, default_vlan=2):
    #TODO: read default vlan from global input config (eg from .virl)
    # TODO: rename to "mark_defaults_vlan" or similar
    # checks all links to managed switches have vlans
    g_vtp = anm['vtp']
    g_l2 = anm['layer2']
    g_l1_conn = anm['layer1_conn']
    managed_switches = [n for n in g_vtp.switches()
                        if n.device_subtype == "managed"]

    no_vlan_ints = []
    for switch in managed_switches:
        for edge in switch.edges():
            neigh_int = edge.dst_int
            local_int = edge.src_int
            if neigh_int.node in managed_switches:
                neigh_int.trunk = True
                continue

            if neigh_int['input'].vlan is None:
                vlan = default_vlan
                no_vlan_ints.append(neigh_int)
            else:
                vlan = neigh_int['input'].vlan

            try:
                vlan = int(vlan)
            except TypeError:
                log.warning("Non-integer vlan %s for %s. Using default %s",
                    vlan, neigh_int, default_vlan)
                vlan = default_vlan

            neigh_int.vlan = vlan
            local_int.vlan = vlan

        for interface in switch:
            if interface.vlan and interface.trunk:
                log.warning("Interface %s set to trunk and vlan", interface)

    # map to layer 2 interfaces
    if len(no_vlan_ints):
        log.info("Setting default VLAN %s to interfaces connected to a managed "
           "switch with no VLAN: %s", default_vlan, no_vlan_ints)


def build_vlans(anm):
    import itertools
    from collections import defaultdict
    g_l2 = anm['layer2']
    g_l1_conn = anm['layer1_conn']
    g_phy = anm['phy']

    g_vtp = anm.add_overlay('vtp')
    # g_vlan = anm.add_overlay('vlan')
    managed_switches = [n for n in g_l2.switches()
                        if n.device_subtype == "managed"]

    g_vtp.add_nodes_from(g_l1_conn)
    g_vtp.add_edges_from(g_l1_conn.edges())

    # remove anything not a managed_switch or connected to a managed_switch
    keep = set()
    keep.update(managed_switches)
    for switch in managed_switches:
        keep.update(switch['vtp'].neighbors())

    remove = set(g_vtp) - keep
    g_vtp.remove_nodes_from(remove)

    edges_to_remove = [e for e in g_vtp.edges()
                       if not(e.src in managed_switches or e.dst in managed_switches)]
    g_vtp.remove_edges_from(edges_to_remove)

    # import ipdb
    # ipdb.set_trace()
    set_default_vlans(anm)

    # copy across vlans from input graph
    vswitch_id_counter = itertools.count(1)

    # TODO: aggregate managed switches

    bcs_to_trim = set()

    subs = ank_utils.connected_subgraphs(g_vtp, managed_switches)
    for sub in subs:
        # identify the VLANs on these switches
        vlans = defaultdict(list)
        sub_neigh_ints = set()
        for switch in sub:
            l2_switch = switch['layer2']
            bcs_to_trim.update(l2_switch.neighbors())
            neigh_ints = {iface for iface in switch.neighbor_interfaces()
                          if iface.node.is_l3device()
                          and iface.node not in sub}

            sub_neigh_ints.update(neigh_ints)

        for interface in sub_neigh_ints:
            # store keyed by vlan id
            vlan = interface['vtp'].vlan
            vlans[vlan].append(interface)

        log.debug("Vlans for sub %s are %s", sub, vlans)
        # create a virtual switch for each
        # TODO: naming: if this is the only pair then name after these, else
        # use the switch names too
        #vswitch_prefix = "_".join(str(sw) for sw in sub)
        vswitches = []  # store to connect trunks
        for vlan, interfaces in vlans.items():
            # create a virtual switch
            vswitch_id = "vswitch%s" % vswitch_id_counter.next()
            # vswitch = g_vlan.add_node(vswitch_id)
            vswitch = g_l2.add_node(vswitch_id)
            # TODO: check of switch or just broadcast_domain for higher layer
            # purposes
            vswitch.device_type = "switch"
            vswitch.device_subtype = "virtual"
            vswitches.append(vswitch)
            # TODO: layout based on midpoint of previous?
            # or if same number as real switches, use their co-ordinates?
            # and then check for coincident?
            vswitch.x = sum(
                i.node['phy'].x for i in interfaces) / len(interfaces) + 50
            vswitch.y = sum(
                i.node['phy'].y for i in interfaces) / len(interfaces) + 50
            vswitch.vlan = vlan

            vswitch['layer2'].broadcast_domain = True
            vswitch['layer2'].vlan = vlan

            # and connect from vswitch to the interfaces
            edges_to_add = [(vswitch, iface) for iface in interfaces]
            g_l2.add_edges_from(edges_to_add)

        # remove the physical switches
        g_l2.remove_nodes_from(bcs_to_trim)
        g_l2.remove_nodes_from(sub)
        # TODO: also remove any broadcast domains no longer connected

        # g_l2.remove_nodes_from(disconnected_bcs)

        # Note: we don't store the interface names as ciuld clobber
        # eg came from two physical switches, each on gige0
        # if need, work backwards from the router iface and its connectivity

        # and add the trunks
        # TODO: these need annotations!
        # create trunks
        edges_to_add = list(itertools.combinations(vswitches, 2))
        # TODO: ensure only once
        # TODO: filter so only one direction
        # g_vlan.add_edges_from(edges_to_add, trunk=True)
        g_vtp.add_edges_from(edges_to_add, trunk=True)



#!/usr/bin/python
# -*- coding: utf-8 -*-
from collections import defaultdict

import autonetkit.log as log
import netaddr
from autonetkit.ank import sn_preflen_to_network
from autonetkit.compiler import sort_sessions
from autonetkit.compilers.device.router_base import RouterCompiler
from autonetkit.nidb import ConfigStanza


class IosBaseCompiler(RouterCompiler):

    """Base IOS compiler"""

    lo_interface_prefix = 'Loopback'
    lo_interface = '%s%s' % (lo_interface_prefix, 0)

    def ibgp_session_data(self, session, ip_version):
        """Wraps RouterCompiler ibgp_session_data
        adds vpnv4 = True if ip_version == 4 and session is in g_ibgp_vpn_v4"""

        data = super(IosBaseCompiler, self).ibgp_session_data(session,
                                                              ip_version)
        if ip_version == 4:
            g_ibgp_vpn_v4 = self.anm['ibgp_vpn_v4']
            if g_ibgp_vpn_v4.has_edge(session):
                data['use_vpnv4'] = True
        return data

    def compile(self, node):
        self.vrf_igp_interfaces(node)
        phy_node = self.anm['phy'].node(node)

        node.use_cdp = phy_node.use_cdp

        if node in self.anm['ospf']:
            node.add_stanza("ospf")
            node.ospf.use_ipv4 = phy_node.use_ipv4
            node.ospf.use_ipv6 = phy_node.use_ipv6

        if node in self.anm['eigrp']:
            node.add_stanza("eigrp")
            node.eigrp.use_ipv4 = phy_node.use_ipv4
            node.eigrp.use_ipv6 = phy_node.use_ipv6

        if node in self.anm['isis']:
            node.add_stanza("isis")
            node.isis.use_ipv4 = phy_node.use_ipv4
            node.isis.use_ipv6 = phy_node.use_ipv6

        if node in self.anm['rip']:
            node.add_stanza("rip")
            node.rip.use_ipv4 = phy_node.use_ipv4
            node.rip.use_ipv6 = phy_node.use_ipv6            

        super(IosBaseCompiler, self).compile(node)
        if node in self.anm['isis']:
            self.isis(node)

        node.label = self.anm['phy'].node(node).label
        self.vrf(node)

    def interfaces(self, node):
        phy_loopback_zero = self.anm['phy'
                                     ].interface(node.loopback_zero)
        if node.ip.use_ipv4:
            ipv4_loopback_subnet = netaddr.IPNetwork('0.0.0.0/32')
            ipv4_loopback_zero = phy_loopback_zero['ipv4']
            ipv4_address = ipv4_loopback_zero.ip_address
            node.loopback_zero.use_ipv4 = True
            node.loopback_zero.ipv4_address = ipv4_address
            node.loopback_zero.ipv4_subnet = ipv4_loopback_subnet
            node.loopback_zero.ipv4_cidr = \
                sn_preflen_to_network(ipv4_address,
                                      ipv4_loopback_subnet.prefixlen)

        if node.ip.use_ipv6:
            # TODO: clean this up so can set on router_base: call cidr not
            # address and update templates
            node.loopback_zero.use_ipv6 = True
            ipv6_loopback_zero = phy_loopback_zero['ipv6']
            node.loopback_zero.ipv6_address = \
                sn_preflen_to_network(ipv6_loopback_zero.ip_address,
                                      128)

        super(IosBaseCompiler, self).interfaces(node)

        for interface in node.physical_interfaces():
            interface.use_cdp = node.use_cdp  # use node value

        for interface in node.interfaces:
            interface.sub_ints = []  # temporary until full subinterfaces

        for interface in node.physical_interfaces():
            g_ext_conn = self.anm['ext_conn']
            if node not in g_ext_conn:
                continue

            node_ext_conn = g_ext_conn.node(node)
            ext_int = node_ext_conn.interface(interface)
            for sub_int in ext_int.sub_int or []:
                stanza = ConfigStanza(
                    id=sub_int['id'],
                    ipv4_address=sub_int['ipv4_address'],
                    ipv4_prefixlen=sub_int['ipv4_prefixlen'],
                    ipv4_subnet=sub_int['ipv4_subnet'],
                    dot1q=sub_int['dot1q'],
                )
                interface.sub_ints.append(stanza)

    def mpls_oam(self, node):
        g_mpls_oam = self.anm['mpls_oam']
        node.add_stanza("mpls")
        node.mpls.oam = False

        if node not in g_mpls_oam:
            return   # no mpls oam configured

        # Node is present in mpls OAM
        node.mpls.oam = True

    def mpls_te(self, node):
        g_mpls_te = self.anm['mpls_te']

        if node not in g_mpls_te:
            return   # no mpls te configured

        node.mpls_te = True

        if node.isis:
            node.isis.ipv4_mpls_te = True
            node.isis.mpls_te_router_id = self.lo_interface

        if node.ospf:
            node.ospf.ipv4_mpls_te = True
            node.ospf.mpls_te_router_id = self.lo_interface

    def nailed_up_routes(self, node):
        log.debug('Configuring nailed up routes')
        phy_node = self.anm['phy'].node(node)

        if node.is_ebgp_v4 and node.ip.use_ipv4:
            infra_blocks = self.anm['ipv4'].data['infra_blocks'
                                                 ].get(phy_node.asn) or []
            for infra_route in infra_blocks:
                stanza = ConfigStanza(
                    prefix=str(infra_route.network),
                    netmask=str(infra_route.netmask),
                    nexthop="Null0",
                    metric=254,
                )
                node.ipv4_static_routes.append(stanza)

        if node.is_ebgp_v6 and node.ip.use_ipv6:
            infra_blocks = self.anm['ipv6'].data['infra_blocks'
                                                 ].get(phy_node.asn) or []
            # TODO: setup schema with defaults
            for infra_route in infra_blocks:
                stanza = ConfigStanza(
                    prefix=str(infra_route),
                    nexthop="Null0",
                    metric=254,
                )
                node.ipv6_static_routes.append(stanza)

    def bgp(self, node):
        node.add_stanza("bgp")
        node.bgp.lo_interface = self.lo_interface
        super(IosBaseCompiler, self).bgp(node)
        phy_node = self.anm['phy'].node(node)
        asn = phy_node.asn
        g_ebgp_v4 = self.anm['ebgp_v4']
        g_ebgp_v6 = self.anm['ebgp_v6']

        if node in g_ebgp_v4 \
                and len(list(g_ebgp_v4.node(node).edges())) > 0:
            node.is_ebgp_v4 = True
        else:
            node.is_ebgp_v4 = False

        if node in g_ebgp_v6 \
                and len(list(g_ebgp_v6.node(node).edges())) > 0:
            node.is_ebgp_v6 = True
        else:
            node.is_ebgp_v6 = False

        node.bgp.ipv4_advertise_subnets = []
        node.bgp.ipv6_advertise_subnets = []

        # Advertise loopbacks into BGP

        if node.ip.use_ipv4:
            node.bgp.ipv4_advertise_subnets = \
                [node.loopback_zero.ipv4_cidr]
        if node.ip.use_ipv6:
            node.bgp.ipv6_advertise_subnets = \
                [node.loopback_zero.ipv6_address]

        # Advertise infrastructure into eBGP

        if node.ip.use_ipv4 and node.is_ebgp_v4:
            infra_blocks = self.anm['ipv4'].data['infra_blocks'
                                                 ].get(asn) or []
            for infra_route in infra_blocks:
                node.bgp.ipv4_advertise_subnets.append(infra_route)

        if node.ip.use_ipv6 and node.is_ebgp_v6:
            infra_blocks = self.anm['ipv6'].data['infra_blocks'
                                                 ].get(asn) or []
            for infra_route in infra_blocks:
                node.bgp.ipv6_advertise_subnets.append(infra_route)

        self.nailed_up_routes(node)

        # vrf
        # TODO: this should be inside vrf section?

        node.bgp.vrfs = []

        vrf_node = self.anm['vrf'].node(node)
        if vrf_node and vrf_node.vrf_role is 'PE':

            # iBGP sessions for this VRF

            vrf_ibgp_neighbors = defaultdict(list)

            g_ibgp_v4 = self.anm['ibgp_v4']
            for session in sort_sessions(g_ibgp_v4.edges(vrf_node)):
                if session.exclude and session.vrf:
                    data = self.ibgp_session_data(session, ip_version=4)
                    stanza = ConfigStanza(data)
                    vrf_ibgp_neighbors[session.vrf].append(stanza)

            g_ibgp_v6 = self.anm['ibgp_v6']
            for session in sort_sessions(g_ibgp_v6.edges(vrf_node)):
                if session.exclude and session.vrf:
                    data = self.ibgp_session_data(session, ip_version=6)
                    stanza = ConfigStanza(data)
                    vrf_ibgp_neighbors[session.vrf].append(stanza)

            # eBGP sessions for this VRF

            vrf_ebgp_neighbors = defaultdict(list)

            for session in sort_sessions(g_ebgp_v4.edges(vrf_node)):
                if session.exclude and session.vrf:
                    data = self.ebgp_session_data(session, ip_version=4)
                    stanza = ConfigStanza(data)
                    vrf_ebgp_neighbors[session.vrf].append(stanza)

            for session in sort_sessions(g_ebgp_v6.edges(vrf_node)):
                if session.exclude and session.vrf:
                    data = self.ebgp_session_data(session, ip_version=6)
                    stanza = ConfigStanza(data)
                    vrf_ebgp_neighbors[session.vrf].append(stanza)

            for vrf in vrf_node.node_vrf_names:
                rd_index = vrf_node.rd_indices[vrf]
                rd = '%s:%s' % (node.asn, rd_index)
                stanza = ConfigStanza(
                    vrf=vrf,
                    rd=rd,
                    use_ipv4=node.ip.use_ipv4,
                    use_ipv6=node.ip.use_ipv6,
                    vrf_ebgp_neighbors=vrf_ebgp_neighbors[vrf],
                    vrf_ibgp_neighbors=vrf_ibgp_neighbors[vrf],
                )
                node.bgp.vrfs.append(stanza)

        # Retain route_target if in ibgp_vpn_v4 and RR or HRR (set in design)

        vpnv4_node = self.anm['ibgp_vpn_v4'].node(node)
        if vpnv4_node:
            retain = False
            if vpnv4_node.retain_route_target:
                retain = True
            node.bgp.vpnv4 = ConfigStanza(retain_route_target=retain)

    def vrf_igp_interfaces(self, node):

        # marks physical interfaces to exclude from IGP

        vrf_node = self.anm['vrf'].node(node)
        if vrf_node and vrf_node.vrf_role is 'PE':
            for interface in node.physical_interfaces():
                vrf_int = self.anm['vrf'].interface(interface)
                if vrf_int.vrf_name:
                    interface.exclude_igp = True

    def vrf(self, node):
        g_vrf = self.anm['vrf']
        vrf_node = self.anm['vrf'].node(node)
        node.add_stanza("vrf")
        node.add_stanza("mpls")
        node.vrf.vrfs = []
        if vrf_node and vrf_node.vrf_role is 'PE':

            # TODO: check if mpls ldp already set elsewhere

            for vrf in vrf_node.node_vrf_names:
                route_target = g_vrf.data.route_targets[node.asn][vrf]
                rd_index = vrf_node.rd_indices[vrf]
                rd = '%s:%s' % (node.asn, rd_index)

                stanza = ConfigStanza(
                    vrf=vrf, rd=rd, route_target=route_target)
                node.vrf.vrfs.append(stanza)

            for interface in node.interfaces:
                vrf_int = self.anm['vrf'].interface(interface)
                if vrf_int.vrf_name:
                    # mark interface as being part of vrf
                    interface.vrf = vrf_int.vrf_name
                    if interface.physical:
                        interface.description += ' vrf %s' \
                            % vrf_int.vrf_name

        if vrf_node and vrf_node.vrf_role in ('P', 'PE'):

            # Add PE -> P, PE -> PE interfaces to MPLS LDP

            node.mpls.ldp_interfaces = []
            for interface in node.physical_interfaces():
                mpls_ldp_int = self.anm['mpls_ldp'].interface(interface)
                if mpls_ldp_int.is_bound:
                    node.mpls.ldp_interfaces.append(interface.id)
                    interface.use_mpls = True

        if vrf_node and vrf_node.vrf_role is 'P':
            node.mpls.ldp_interfaces = []
            for interface in node.physical_interfaces():
                node.mpls.ldp_interfaces.append(interface.id)

        vrf_node = self.anm['vrf'].node(node)

        node.vrf.use_ipv4 = node.ip.use_ipv4
        node.vrf.use_ipv6 = node.ip.use_ipv6
        node.vrf.vrfs = sorted(node.vrf.vrfs, key=lambda x: x.vrf)

        if self.anm.has_overlay('mpls_ldp') and node \
                in self.anm['mpls_ldp']:
            node.mpls.enabled = True
            node.mpls.router_id = node.loopback_zero.id

    def rip(self, node):
    #Inheriting from base compiler. Adding in interface stanza.
        super(IosBaseCompiler, self).rip(node)
        for interface in node.physical_interfaces():
            phy_int = self.anm['phy'].interface(interface)

            rip_int = phy_int['rip']
            if rip_int and rip_int.is_bound:
                if interface.exclude_igp:
                    continue  

                interface.rip = {
                    'cost': rip_int.cost,
                    'area': rip_int.area,
                    'process_id': node.rip.process_id,
                    'use_ipv4': node.ip.use_ipv4,
                    'use_ipv6': node.ip.use_ipv6,
                }

    def ospf(self, node):
        super(IosBaseCompiler, self).ospf(node)
        for interface in node.physical_interfaces():
            phy_int = self.anm['phy'].interface(interface)

            ospf_int = phy_int['ospf']
            if ospf_int and ospf_int.is_bound:
                if interface.exclude_igp:
                    continue  # don't configure IGP for this interface

                # TODO: use ConfigStanza here
                interface.ospf = {
                    'cost': ospf_int.cost,
                    'area': ospf_int.area,
                    'process_id': node.ospf.process_id,
                    'use_ipv4': node.ip.use_ipv4,
                    'use_ipv6': node.ip.use_ipv6,
                    'multipoint': ospf_int.multipoint,
                }

                # TODO: add wrapper for this

    def eigrp(self, node):
        super(IosBaseCompiler, self).eigrp(node)
        for interface in node.physical_interfaces():
            phy_int = self.anm['phy'].interface(interface)

            eigrp_int = phy_int['eigrp']
            if eigrp_int and eigrp_int.is_bound:
                if interface.exclude_igp:
                    continue  # don't configure IGP for this interface

                interface.eigrp = {
                    'metric': eigrp_int.metric,
                    'area': eigrp_int.area,
                    'name': node.eigrp.name,
                    'use_ipv4': node.ip.use_ipv4,
                    'use_ipv6': node.ip.use_ipv6,
                    'multipoint': eigrp_int.multipoint,
                }

                # TODO: add wrapper for this

    def isis(self, node):
        super(IosBaseCompiler, self).isis(node)
        for interface in node.physical_interfaces():
            isis_int = self.anm['isis'].interface(interface)
            edges = isis_int.edges()
            if not isis_int.is_bound:
                # Could occur for VRFs
                log.debug("No ISIS connections for interface %s" % interface)
                continue

            # TODO: change this to be is_bound and is_multipoint
            if isis_int.multipoint:
                log.warning('Extended IOS config support not valid for multipoint ISIS connections on %s'
                            % interface)
                continue

                # TODO multipoint handling?

            edge = edges[0]
            dst = edge.dst
            if not dst.is_router():
                log.debug('Connection to non-router host not added to IGP'
                          )
                continue

            src_type = node.device_subtype
            dst_type = dst['phy'].device_subtype
            if src_type == 'IOS XRv':
                if dst_type == 'IOSv':
                    interface.isis.hello_padding_disable = True
                elif dst_type == 'CSR1000v':
                    interface.isis.hello_padding_disable = True
                elif dst_type == 'NX-OSv':
                    interface.isis.hello_padding_disable = True

            if src_type == 'IOSv':
                if dst_type == 'IOS XRv':
                    interface.isis.mtu = 1430

            if src_type == 'CSR1000v':
                if dst_type == 'IOS XRv':
                    interface.isis.mtu = 1430

            if src_type == 'NX-OSv':
                if dst_type == 'IOS XRv':
                    interface.mtu = 1430  # for all of interface
                    interface.isis.hello_padding_disable = True
                elif dst_type == 'IOSv':
                    interface.isis.hello_padding_disable = True
                elif dst_type == 'CSR1000v':
                    interface.isis.hello_padding_disable = True

            interface.isis_mtu = interface.isis.mtu
            interface.hello_padding_disable = \
                interface.isis.hello_padding_disable


class IosClassicCompiler(IosBaseCompiler):

    def compile(self, node):
        super(IosClassicCompiler, self).compile(node)

        self.mpls_te(node)
        self.mpls_oam(node)
        self.gre(node)
        self.l2tp_v3(node)

        phy_node = self.anm['phy'].node(node)
        if phy_node.device_subtype == 'IOSv':

            # only copy across for certain reference platforms

            node.use_onepk = phy_node.use_onepk
            node.transport_input_ssh_telnet = True
            node.no_service_config = True
            node.ipv4_cef = True
            node.ipv6_cef = True

        if phy_node.device_subtype == 'CSR1000v':

            # only copy across for certain reference platforms

            node.use_onepk = phy_node.use_onepk
            node.transport_input_ssh_telnet = True
            node.include_csr = True

            # Set secret password to "cisco"

            node.enable_secret = \
                'tnhtc92DXBhelxjYk8LWJrPV36S2i4ntXrpb4RFmfqY'
            node.exclude_phy_int_auto_speed_duplex = True
            node.no_service_config = True

    def rip(self, node):
        super(IosClassicCompiler, self).rip(node)

    def ospf(self, node):
        super(IosClassicCompiler, self).ospf(node)
        loopback_zero = node.loopback_zero
        ospf_node = self.anm['ospf'].node(node)
        loopback_zero.ospf = {
            'cost': 1,
            'area': ospf_node.area,
            'process_id': node.ospf.process_id,
            'use_ipv4': False,
            'use_ipv6': node.ip.use_ipv6,
            'multipoint': False,
        }

        # TODO: add wrapper for this
    def gre(self, node):
        node.gre_tunnels = []
        if not self.anm.has_overlay('gre_tunnel'):
            return

        g_gre_tunnel = self.anm['gre_tunnel']
        if node not in g_gre_tunnel:
            return   # no gre tunnel for node

        gre_node = g_gre_tunnel.node(node)
        neighbors = gre_node.neighbors()
        for index, neigh in enumerate(neighbors, start=1):
            stanza = ConfigStanza(id=index, endpoint=neigh)

            # TODO: try/except here
            # TODO: Explain logic here
            src_int = g_gre_tunnel.edge(node, neigh).src_int
            tunnel_source = node.interface(src_int).id
            stanza.source = tunnel_source
            stanza.destination = "0.0.0.0"  # placeholder for user to replace

            if neigh.tunnel_enabled_ipv4:
                ip_address = neigh.tunnel_ipv4_address
                cidr = neigh.tunnel_ipv4_cidr
                stanza.ipv4_address = ip_address
                stanza.ipv4_subnet = cidr
                stanza.use_ipv4 = True

            if neigh.tunnel_enabled_ipv6:
                cidr = neigh.tunnel_ipv6_cidr
                stanza.ipv4_subnet = cidr
                stanza.use_ipv6 = True

            node.gre_tunnels.append(stanza)

    def l2tp_v3(self, node):
        node.l2tp_classes = []
        node.pseudowire_classes = []
        if not self.anm.has_overlay('l2tp_v3'):
            return

        g_l2tp_v3 = self.anm['l2tp_v3']
        g_phy = self.anm['phy']
        if node not in g_l2tp_v3:
            return   # no l2tp_v3 for node

        l2tp_v3_node = g_l2tp_v3.node(node)
        if l2tp_v3_node.role != "tunnel":
            return # nothing to configure

        node.l2tp_classes = list(l2tp_v3_node.l2tp_classes)
        for l2tp_class in l2tp_v3_node.l2tp_classes:
            node.l2tp_classes

        node.pseudowire_classes = []
        for pwc in l2tp_v3_node.pseudowire_classes:
            stanza = ConfigStanza()
            stanza.name = pwc['name']
            stanza.encapsulation = pwc['encapsulation']
            stanza.protocol = pwc['protocol']
            stanza.l2tp_class_name = pwc['l2tp_class_name']
            local_interface = pwc['local_interface']

            # Lookup the interface ID allocated for this loopback by compiler
            local_interface_id = node.interface(local_interface).id
            stanza.local_interface = local_interface_id

            node.pseudowire_classes.append(stanza)

        for interface in node.physical_interfaces():
            phy_int = g_phy.interface(interface)
            if phy_int.xconnect_encapsulation != "l2tpv3":
                continue # no l2tpv3 encap, no need to do anything

            tunnel_int = l2tp_v3_node.interface(interface)
            stanza = ConfigStanza()
            stanza.remote_ip = tunnel_int.xconnect_remote_ip
            stanza.vc_id = tunnel_int.xconnect_vc_id
            stanza.encapsulation = "l2tpv3"
            #TODO: need to be conscious of support for other xconnect types
            # in templates since pw_class may not apply if not l2tpv3, et
            stanza.pw_class = tunnel_int.xconnect_pw_class

            interface.xconnect = stanza

    def eigrp(self, node):
        super(IosClassicCompiler, self).eigrp(node)
        # Numeric process IDs use "old-style" non-ipv6 EIGRP stanzas
        process_id = node.eigrp.process_id
        if str(process_id).isdigit():
            process_id = "as%s" % process_id
        node.eigrp.process_id = process_id

    def mpls_te(self, node):
        super(IosClassicCompiler, self).mpls_te(node)

        g_mpls_te = self.anm['mpls_te']
        if node not in g_mpls_te:
            return   # no mpls te configured

        mpls_te_node = g_mpls_te.node(node)

        for interface in mpls_te_node.physical_interfaces():
            nidb_interface = self.nidb.interface(interface)
            if not interface.is_bound:
                log.debug('Not enable MPLS and RSVP for interface %s on %s '
                          % (nidb_interface.id, node))
                continue
            nidb_interface.te_tunnels = True
            nidb_interface.rsvp_bandwidth_percent = 100

    def bgp(self, node):
        super(IosClassicCompiler, self).bgp(node)

        node.bgp.use_ipv4 = node.ip.use_ipv4
        node.bgp.use_ipv6 = node.ip.use_ipv6

        # Seperate by address family

        ipv4_peers = []
        ipv6_peers = []

        # Note cast to dict - #TODO revisit this requirement
        # TODO: revisit and tidy up the logic here: split iBGP and eBGP
        # TODO: sort the peer list by peer IP

        for peer in node.bgp.ibgp_neighbors:
            peer = ConfigStanza(peer)
            peer.remote_ip = peer.loopback
            if peer.use_ipv4:
                if node.is_ebgp_v4:
                    peer.next_hop_self = True
                ipv4_peers.append(peer)
            if peer.use_ipv6:
                if node.is_ebgp_v6:
                    peer.next_hop_self = True
                ipv6_peers.append(peer)

        for peer in node.bgp.ibgp_rr_parents:
            peer = ConfigStanza(peer)
            peer.remote_ip = peer.loopback
            if peer.use_ipv4:
                if node.is_ebgp_v4:
                    peer.next_hop_self = True
                ipv4_peers.append(peer)
            if peer.use_ipv6:
                if node.is_ebgp_v6:
                    peer.next_hop_self = True
                ipv6_peers.append(peer)

        for peer in node.bgp.ibgp_rr_clients:
            peer = ConfigStanza(peer)
            peer.rr_client = True
            peer.remote_ip = peer.loopback
            if peer.use_ipv4:
                if node.is_ebgp_v4:
                    peer.next_hop_self = True
                ipv4_peers.append(peer)
            if peer.use_ipv6:
                if node.is_ebgp_v6:
                    peer.next_hop_self = True
                ipv6_peers.append(peer)

        for peer in node.bgp.ebgp_neighbors:
            peer = ConfigStanza(peer)
            peer.is_ebgp = True
            peer.remote_ip = peer.dst_int_ip
            if peer.use_ipv4:
                peer.next_hop_self = True
                ipv4_peers.append(peer)
            if peer.use_ipv6:
                peer.next_hop_self = True
                ipv6_peers.append(peer)

        node.bgp.ipv4_peers = ipv4_peers
        node.bgp.ipv6_peers = ipv6_peers

        vpnv4_neighbors = []
        if node.bgp.vpnv4:
            for neigh in node.bgp.ibgp_neighbors:
                if not neigh.use_ipv4:
                    continue

                neigh_data = ConfigStanza(neigh)
                vpnv4_neighbors.append(neigh_data)

            for neigh in node.bgp.ibgp_rr_clients:
                if not neigh.use_ipv4:
                    continue
                neigh_data = ConfigStanza(neigh)
                neigh_data.rr_client = True
                vpnv4_neighbors.append(neigh_data)

            for neigh in node.bgp.ibgp_rr_parents:
                if not neigh.use_ipv4:
                    continue
                neigh_data = ConfigStanza(neigh)
                vpnv4_neighbors.append(neigh_data)

        vpnv4_neighbors = sorted(vpnv4_neighbors, key=lambda x:
                                 x['loopback'])
        node.bgp.vpnv4_neighbors = vpnv4_neighbors


class IosXrCompiler(IosBaseCompiler):

    def compile(self, node):
        super(IosXrCompiler, self).compile(node)
        self.mpls_te(node)
        self.mpls_oam(node)

    def mpls_te(self, node):
        super(IosXrCompiler, self).mpls_te(node)

        g_mpls_te = self.anm['mpls_te']
        if node not in g_mpls_te:
            return   # no mpls te configured

        rsvp_interfaces = []
        mpls_te_interfaces = []
        mpls_te_node = g_mpls_te.node(node)

        for interface in mpls_te_node.physical_interfaces():
            nidb_interface = self.nidb.interface(interface)
            stanza = ConfigStanza(id=nidb_interface.id,
                                  bandwidth_percent=100)
            rsvp_interfaces.append(stanza)

            mpls_te_interfaces.append(nidb_interface.id)

        node.add_stanza("rsvp")
        node.rsvp.interfaces = rsvp_interfaces
        node.mpls.te_interfaces = mpls_te_interfaces
   
    def rip(self, node):
        super(IosXrCompiler, self).rip(node)

        g_rip = self.anm['rip']
        ipv4_interfaces = []

        for interface in node.physical_interfaces():
            if interface.exclude_igp:
                continue  # discontinue configuring IGP for this interface

            rip_int = g_rip.interface(interface)
            if rip_int and rip_int.is_bound:
                data = {'id': interface.id, 'passive': False}
                stanza = ConfigStanza(**data)
                if node.rip.use_ipv4:
                    ipv4_interfaces.append(stanza)

        loopback_zero = node.loopback_zero
        data = {'id': node.loopback_zero.id, 'passive': True}
        stanza = ConfigStanza(**data)
        if node.rip.use_ipv4:
            ipv4_interfaces.append(stanza)


        node.rip.ipv4_interfaces = ipv4_interfaces

    def ospf(self, node):
        super(IosXrCompiler, self).ospf(node)

    def eigrp(self, node):
        super(IosXrCompiler, self).eigrp(node)

        g_eigrp = self.anm['eigrp']
        ipv4_interfaces = []
        ipv6_interfaces = []

        for interface in node.physical_interfaces():
            if interface.exclude_igp:
                continue  # don't configure IGP for this interface

            eigrp_int = g_eigrp.interface(interface)
            if eigrp_int and eigrp_int.is_bound:
                # TODO: for here and below use stanza directly
                data = {'id': interface.id, 'passive': False}
                stanza = ConfigStanza(**data)
                if node.eigrp.use_ipv4:
                    ipv4_interfaces.append(stanza)
                if node.eigrp.use_ipv6:
                    ipv6_interfaces.append(stanza)

        loopback_zero = node.loopback_zero
        data = {'id': node.loopback_zero.id, 'passive': True}
        stanza = ConfigStanza(**data)
        if node.eigrp.use_ipv4:
            ipv4_interfaces.append(stanza)
        if node.eigrp.use_ipv6:
            ipv6_interfaces.append(stanza)

        node.eigrp.ipv4_interfaces = ipv4_interfaces
        node.eigrp.ipv6_interfaces = ipv6_interfaces

    def isis(self, node):
        super(IosXrCompiler, self).isis(node)
        node.isis.isis_links = []

        for interface in node.physical_interfaces():
            if interface.exclude_igp:
                continue  # don't configure IGP for this interface

            # print interface.isis.dump()
            # copy across attributes from the IosBaseCompiler setting step

            isis_int = self.anm['isis'].interface(interface)
            if isis_int and isis_int.is_bound:
                data = {'id': interface.id, 'metric': isis_int.metric,
                        'multipoint': isis_int.multipoint}
                if interface.isis.hello_padding_disable is not None:
                    data['hello_padding_disable'] = \
                        interface.isis.hello_padding_disable
                if interface.isis.mtu is not None:
                    data['mtu'] = interface.isis.hello_padding_disable

                # TODO: make stanza
                stanza = ConfigStanza(**data)
                node.isis.isis_links.append(stanza)


class NxOsCompiler(IosBaseCompiler):

    def compile(self, node):
        super(NxOsCompiler, self).compile(node)
        self.mpls_te(node)
        self.mpls_oam(node)

    def mpls_te(self, node):
        g_mpls_te = self.anm['mpls_te']
        if node not in g_mpls_te:
            return   # no mpls te configured

        if node.supported_features.mpls_te is False:
            node.log.warning("Feature MPLS TE is not supported for %s on the %s platform" % (
                node.device_subtype, node.platform))

    def mpls_oam(self, node):
        g_mpls_oam = self.anm['mpls_oam']
        if node not in g_mpls_oam:
            return   # no mpls oam configured

        if node.supported_features.mpls_oam is False:
            node.log.warning("Feature MPLS OAM is not supported for %s the on %s platform" % (
                node.device_subtype, node.platform))

    def vrf(self, node):
        g_vrf = self.anm['vrf']
        if node not in g_vrf:
            return   # no mpls oam configured
        if node.supported_features.vrf is False:
            node.log.warning("Feature VRF is not supported for %s on the %s platform" % (
                node.device_subtype, node.platform))

    def interfaces(self, node):

        # need to aggregate areas

        super(NxOsCompiler, self).interfaces(node)

    def rip(self, node):
        super(NxOsCompiler, self).rip(node)
        loopback_zero = node.loopback_zero
        loopback_zero.rip = {'use_ipv4': node.ip.use_ipv4,
                            'process_id': node.rip.process_id}     


    def ospf(self, node):
        super(NxOsCompiler, self).ospf(node)
        loopback_zero = node.loopback_zero
        g_ospf = self.anm['ospf']
        ospf_node = g_ospf.node(node)
        loopback_zero.ospf = {
            'cost': ospf_node.cost,
            'area': ospf_node.area,
            'process_id': node.ospf.process_id,
            'use_ipv4': node.ip.use_ipv4,
            'use_ipv6': node.ip.use_ipv6,
        }

        # TODO: add wrapper for this

    def eigrp(self, node):

        # TODO: do we want to specify the name or hard-code (as currently)?

        super(NxOsCompiler, self).eigrp(node)
        loopback_zero = node.loopback_zero
        loopback_zero.eigrp = {'use_ipv4': node.ip.use_ipv4,
                               'use_ipv6': node.ip.use_ipv6}


class StarOsCompiler(IosBaseCompiler):

    pass



import logging
import logging.handlers
import autonetkit.config as config

ank_logger = logging.getLogger("ANK2")
if not ank_logger.handlers:
    console_formatter = logging.Formatter("%(levelname)-1s %(type)s %(id)s %(message)s")
    ch = logging.StreamHandler()
    #ch.setLevel(logging.INFO)
    ch.setFormatter(console_formatter)
    ch.setLevel(logging.INFO)
    ank_logger.addHandler(ch)



ank_logger.setLevel(logging.INFO)
# Reference for external access
logger = ank_logger
# Use approach of Pika, allows for autonetkit.log.debug("message")
debug = logger.debug
error = logger.error
info = logger.info
warning = logger.warning
exception = logger.exception
critical = logger.critical



import autonetkit.log as log
from autonetkit.anm.base import OverlayBase


class OverlaySubgraph(OverlayBase):

    """OverlaySubgraph"""

    def __init__(self, anm, overlay_id, graph, name=None):
        """"""

        object.__setattr__(self, '_subgraph_name', name)
        super(OverlaySubgraph, self).__init__(anm, overlay_id)
        self._graph = graph

    def __repr__(self):
        return self._subgraph_name or 'subgraph'



# based on
# http://reminiscential.wordpress.com/2012/04/07/realtime-notification-delivery-using-rabbitmq-tornado-and-websocket/
import json
import logging
import os
import socket

import autonetkit.config as config
import pkg_resources
import tornado
import tornado.websocket as websocket


class MyWebHandler(tornado.web.RequestHandler):

    def initialize(self, ank_accessor, singleuser_mode=False):
        self.ank_accessor = ank_accessor
        self.singleuser_mode = singleuser_mode

    def get(self):
        # TODO: make thie echo the ank version, and suggest how to post
        # topologies
        self.write("Hello, world")

    def post(self):
        data = self.get_argument('body', 'No data received')
        data_type = self.get_argument('type', 'No data received')
        if self.singleuser_mode:
            uuid = "singleuser"
        else:
            uuid = self.get_argument('uuid', 'singleuser')

        # get listeners for this uuid
        uuid_socket_listeners = self.application.socket_listeners[uuid]

        if data_type == "anm":
            body_parsed = json.loads(data)

            # TODO: if single user mode then fall back to single user

            if False:  # use to save the default.json
                import gzip
                vis_content = pkg_resources.resource_filename(
                    "autonetkit_vis", "web_content")
                with gzip.open(os.path.join(vis_content, "default.json.gz"), "w") as fh:
                    json.dump(body_parsed, fh)

            self.ank_accessor.store_overlay(uuid, body_parsed)
            logging.info("Updating listeners")
            for listener in uuid_socket_listeners:
                listener.update_overlay()

        elif data_type == "highlight":
            body_parsed = json.loads(data)
            for listener in uuid_socket_listeners:
                listener.write_message({'highlight': body_parsed})
        else:
            pass


class MyWebSocketHandler(websocket.WebSocketHandler):

    def initialize(self, ank_accessor, overlay_id, singleuser_mode=False):
        """ Store the overlay_id this listener is currently viewing.
        Used when updating."""
        self.ank_accessor = ank_accessor
        self.overlay_id = overlay_id
        self.uuid = None  # set by the client
        self.uuid_socket_listeners = set()
        self.singleuser_mode = singleuser_mode

    def allow_draft76(self):
        # for iOS 5.0 Safari
        return True

    def open(self, *args, **kwargs):
        # Tornado needs default (here is None) or throws exception that
        # required argument not provided
        if self.singleuser_mode:
            uuid = "singleuser"
        else:
            uuid = self.get_argument("uuid", "singleuser")

        self.uuid = uuid

        self.uuid_socket_listeners = self.application.socket_listeners[uuid]

        logging.info("Client connected from %s" % self.request.remote_ip)
        self.uuid_socket_listeners.add(self)

    def on_close(self):
        self.uuid_socket_listeners.remove(self)
        logging.info("Client disconnected from %s" % self.request.remote_ip)
        try:
            self.application.pc.remove_event_listener(self)
        except AttributeError:
            pass  # no RabbitMQ server
        try:
            self.application.echo_server.remove_event_listener(self)
        except AttributeError:
            pass  # no echo_server

    def on_message(self, message):
        logging.info("Received message %s from websocket client" % message)
        if "overlay_id" in message:
            # TODO: form JSON on client side, use loads here
            _, overlay_id = message.split("=")
            self.overlay_id = overlay_id
            self.update_overlay()
        elif "overlay_list" in message:
            body = json.dumps(
                {'overlay_list': self.ank_accessor.overlay_list(self.uuid)})
            self.write_message(body)
        elif "ip_allocations" in message:
            pass

    def update_overlay(self):
        body = self.ank_accessor.get_overlay(self.uuid, self.overlay_id)
        self.write_message(body)
        body = json.dumps(
            {'overlay_list': self.ank_accessor.overlay_list(self.uuid)})
        self.write_message(body)


class AnkAccessor():

    """ Used to store published topologies"""

    def __init__(self, maxlen=25, simplified_overlays=False):
        from collections import deque
        self.anm_index = {}
        self.uuid_list = deque(maxlen=maxlen)  # use for circular buffer
        self.anm = {}
        self.ip_allocation = {}
        self.simplified_overlays = simplified_overlays
# try loading from vis directory
        try:
            import autonetkit_cisco_webui
            default_file = pkg_resources.resource_filename(
                "autonetkit_cisco_webui", "cisco.json.gz")
        except ImportError:
            vis_content = pkg_resources.resource_filename(
                "autonetkit_vis", "web_content")
            default_file = os.path.join(vis_content, "default.json.gz")

        try:
            import gzip
            fh = gzip.open(default_file, "r")
            data = json.load(fh)
            #data = json.loads(loaded)
            self.anm_index['singleuser'] = data
        except IOError, e:
            logging.warning(e)
            pass  # use default blank anm

    def store_overlay(self, uuid, overlay_input):
        logging.info("Storing overlay_input with UUID %s" % uuid)

        if self.simplified_overlays:
            overlays_tidied = {}

            overlay_keys = [
                index for index, data in overlay_input.items() if len(data.get("nodes"))]

            keys_to_exclude = {"input",
                               "bgp", "ibgp", "ebgp",
                               "graphics", "ip", "nidb"}
            overlay_keys = [
                k for k in overlay_keys if k not in keys_to_exclude]

            labels = {
                "ibgp_v6": "iBGP v6",
                "ibgp_v4": "iBGP v4",
                "ibgp_vpn_v4": "iBGP VPN v4",
                "ebgp_v6": "eBGP v6",
                "mpls_te": "MPLS TE",
                "mpls_ldp": "MPLS LDP",
                "ebgp_v4": "eBGP v4",
                "mpls_oam": "MPLS OAM",
                "ipv4": "IP v4",
                "ipv6": "IP v6",
                "segment_routing": "Segment Routing",
                "pce": "PCE",
                "bgp_ls": "BGP LS",
                "phy": "Physical",
                "vrf": "VRF",
                "isis": "IS-IS",
                "bgp": "BGP",
                "eigrp": "EIGRP",
                "ebgp": "eBGP",
                "ospf": "OSPF",
            }

            # DIsable until all web engines are verified to support format (eg
            # ank_cisco_webui)
            labels = {}

            # Check if new uuid or updating previous uuid
            for key in overlay_keys:
                # use from labels if present
                store_key = labels.get(key) or key
                overlays_tidied[store_key] = overlay_input[key]

        else:
            overlays_tidied = overlay_input

        # New uuid
        if len(self.uuid_list) == self.uuid_list.maxlen:
            # list is full
            oldest_uuid = self.uuid_list.popleft()
            logging.info("UUID list full, removing UUID %s" % oldest_uuid)

            try:
                del self.anm_index[oldest_uuid]
            except KeyError:
                logging.warning("Unable to remove UUID %s" % oldest_uuid)

        # If uuid already present, then remove from the queue, and then add to the end
        # This avoids erroneously removing recently updated (i.e. non-stale
        # uuids)
        if uuid in self.uuid_list:
            logging.info("Removing UUID %s to add to end of queue" % uuid)
            self.uuid_list.remove(uuid)

        self.uuid_list.append(uuid)
        logging.info("Stored overlay with UUID %s" % uuid)
        self.anm_index[uuid] = overlays_tidied

    def get_overlay(self, uuid, overlay_id):
        logging.info("Getting overlay %s with UUID %s" % (overlay_id, uuid))
        try:
            anm = self.anm_index[uuid]
        except KeyError:
            logging.warning("Unable to find topology with UUID %s" % uuid)
            return ""
        else:
            try:
                if overlay_id == "*":
                    return anm
                # elif self.simplified_overlays and overlay_id == "phy":
                    # print "Returning physical for phy"
                    # return anm["Physical"]
                else:
                    return anm[overlay_id]
            except KeyError:
                logging.warning(
                    "Unable to find overlay %s in topoplogy with UUID %s" % (overlay_id, uuid))

    def overlay_list(self, uuid):
        logging.info("Trying for anm list with UUID %s" % uuid)
        try:
            anm = self.anm_index[uuid]
        except KeyError:
            logging.warning("Unable to find topology with UUID %s" % uuid)
            return [""]

        if not len(anm):
            return [""]

        return sorted(anm.keys(), key=lambda x: str(x[0]).lower())

    def __getitem__(self, key):
        try:
            return self.anm[key]
        except KeyError:
            return json.dumps(["No ANM loaded"])

    def ip_allocations(self):
        return self.ip_allocation


class IndexHandler(tornado.web.RequestHandler):

    """Used to treat index.html as a template and substitute the uuid parameter for the websocket call
    """

    def initialize(self, path):
        self.content_path = path

    def get(self):
        # if not set, use default uuid of "singleuser"
        uuid = self.get_argument("uuid", "singleuser")
        # Note agent detection is usually *not* the solution to the problem, see:
        # https://developer.mozilla.org/en-US/docs/Browser_detection_using_the_user_agent
        user_agent = self.request.headers.get("User-Agent")
        formatted_url = "http://%s/?uuid=%s" % (self.request.host, uuid)
        logging.info(user_agent)
        supported_browsers = ", ".join(["Google Chrome", "Mozilla Firefox"])
        unsupported_browsers = ["MSIE 5.0", "MSIE 6.0", "MSIE 7.0",
                                # "Chrome", # testing
                                #"Mozilla",
                                "MSIE 8.0", "MSIE 9.0", "MSIE 10.0", "MSIE 11.0"]
        if any(x in user_agent for x in unsupported_browsers):
            # bad browser
            fallback_template = ("""Your browser is not supported.<br />
                Please try the following link in a supported browser: <br />
                <a href="%s">%s</a> </br>
                <p>
                Supported browsers are %s.
                </p>
                <br />
                """ % (formatted_url, formatted_url, supported_browsers))

            template = os.path.join(
                self.content_path, "unsupported_browser.html")
            self.render(template, formatted_url=formatted_url,
                        supported_browsers=supported_browsers)
            return

        logging.info("Rendering template with uuid %s" % uuid)
        template = os.path.join(self.content_path, "index.html")
        self.render(template, uuid=uuid)


def main():

    try:
        ANK_VERSION = pkg_resources.get_distribution("autonetkit").version
    except pkg_resources.DistributionNotFound:
        ANK_VERSION = "dev"

    import argparse
    usage = "ank_webserver"
    version = "%(prog)s using AutoNetkit " + str(ANK_VERSION)
    parser = argparse.ArgumentParser(description=usage, version=version)
    parser.add_argument(
        '--port', type=int,  help="Port to run webserver on (default 8000)")
    parser.add_argument(
        '--multi_user', action="store_true", default=False, help="Multi-User mode")
    parser.add_argument('--ank_vis', action="store_true",
                        default=False, help="Force AutoNetkit visualisation system")
    arguments = parser.parse_args()

# check if most recent outdates current most recent
    content_path = None

    try:
        import autonetkit_vis
    except ImportError:
        pass  # #TODO: logging no vis
    else:
        # use web content from autonetkit_cisco module
        content_path = pkg_resources.resource_filename(
            "autonetkit_vis", "web_content")

    # arguments.ank_vis = False # manually force for now

    simplified_overlays = False
    if not arguments.ank_vis:
        try:
            import autonetkit_cisco_webui
            simplified_overlays = True
        except ImportError:
            pass  # use AutoNetkit internal web content
        else:
            # use web content from autonetkit_cisco module
            content_path = pkg_resources.resource_filename(
                "autonetkit_cisco_webui", "web_content")

    if not content_path:
        logging.warning(
            "No visualisation pages found: did you mean to install autonetkit_vis? Exiting...")
        raise SystemExit

    settings = {
        "static_path": content_path,
        'debug': False,
        # otherwise content with folder /static won't get mapped
        "static_url_prefix": "unused",
    }

    singleuser_mode = False  # default for now
    if arguments.multi_user:
        singleuser_mode = False

    if singleuser_mode:
        logging.info("Running webserver in single-user mode")

    ank_accessor = AnkAccessor(simplified_overlays=simplified_overlays)

    # TODO: inherit the IndexHandler to switch based on browser version
    application = tornado.web.Application([
        (r'/ws', MyWebSocketHandler, {"ank_accessor": ank_accessor,
                                      'singleuser_mode': singleuser_mode,
                                      "overlay_id": "phy"}),
        (r'/publish', MyWebHandler, {"ank_accessor": ank_accessor,
                                     'singleuser_mode': singleuser_mode,
                                     }),

        # TODO: merge the two below into a single handler that captures both
        # cases
        (r'/', IndexHandler, {"path": settings['static_path']}),
        (r'/index.html', IndexHandler, {"path": settings['static_path']}),
        ("/(.*)", tornado.web.StaticFileHandler,
         {"path": settings['static_path']})
    ], **settings)

    logging.getLogger().setLevel(logging.INFO)

    from collections import defaultdict
    application.socket_listeners = defaultdict(set)  # Indexed by uuid

    io_loop = tornado.ioloop.IOLoop.instance()

    port = config.settings['Http Post']['port']
    if arguments.port:
        port = arguments.port  # explicitly set on command line

    import time
    timestamp = time.strftime("%Y %m %d_%H:%M:%S", time.localtime())
    logging.info("Starting on port %s at %s" % (port, timestamp))

    try:
        application.listen(port)
    except socket.error, e:
        if e.errno is 48:  # socket in use
            logging.warning(
                "Unable to start webserver: socket in use for port %s" % port)
            raise SystemExit

    io_loop.start()

if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-
import functools
import itertools
import pprint
import string
import time
from functools import total_ordering

import autonetkit.log as log
import networkx as nx
from autonetkit.ank_utils import unwrap_edges, unwrap_nodes

try:
    import cPickle as pickle
except ImportError:
    import pickle

import logging


#TODO: rename duplicate use of logger var in log setup
class CustomAdapter(logging.LoggerAdapter):
    def process(self, msg, kwargs):
        return '[%s]: %s' % (self.extra['item'], msg), kwargs

class AutoNetkitException(Exception):

    pass




# TODO: rename to NmPort





def multi_edge():
    """Returns anm with  input and physical as house graph"""
    from networkx.readwrite import json_graph
    import networkx as nx
    import autonetkit
    # returns a house graph
    data = {'directed': False,
            'graph': [],
            'links': [{'_ports': {'r4': 2, 'r5': 1},
                       'raw_interfaces': {},
                       'source': 0,
                       'target': 1},
                      {'_ports': {'r2': 3, 'r4': 1},
                       'raw_interfaces': {},
                       'source': 0,
                       'target': 3},
                       {'_ports': {'r2': 4, 'r4': 3},
                        'raw_interfaces': {},
                        'source': 0,
                        'target': 3},
                      {'_ports': {'r3': 3, 'r5': 2},
                       'raw_interfaces': {},
                       'source': 1,
                       'target': 4},
                      {'_ports': {'r1': 1, 'r2': 1},
                       'raw_interfaces': {},
                       'source': 2,
                       'target': 3},
                       {'_ports': {'r1': 3, 'r2': 5},
                        'raw_interfaces': {},
                        'source': 2,
                        'target': 3},
                      {'_ports': {'r1': 2, 'r3': 1},
                       'raw_interfaces': {},
                       'source': 2,
                       'target': 4},
                       {'_ports': {'r1': 4, 'r3': 4},
                        'raw_interfaces': {},
                        'source': 2,
                        'target': 4},
                        {'_ports': {'r1': 5, 'r3': 5},
                         'raw_interfaces': {},
                         'source': 2,
                         'target': 4},
                      {'_ports': {'r2': 2, 'r3': 2},
                       'raw_interfaces': {},
                       'source': 3,
                       'target': 4}],
            'multigraph': True,
            'nodes': [{'_ports': {0: {'category': 'physical', 'description': None},
                                  1: {'category': 'physical', 'description': 'r4 to r2', 'id': 'eth0'},
                                  2: {'category': 'physical', 'description': 'r4 to r5', 'id': 'eth1'},
                                  3: {'category': 'physical', 'description': 'r4 to r2', 'id': 'eth2'}},
                       'asn': 2,
                       'device_type': 'router',
                       'id': 'r4',
                       'label': 'r4',
                       'x': 675,
                       'y': 300},
                      {'_ports': {0: {'category': 'physical', 'description': None},
                                  1: {'category': 'physical', 'description': 'r5 to r4', 'id': 'eth0'},
                                  2: {'category': 'physical', 'description': 'r5 to r3', 'id': 'eth1'}},
                       'asn': 2,
                       'device_type': 'router',
                       'id': 'r5',
                       'label': 'r5',
                       'x': 675,
                       'y': 500},
                      {'_ports': {0: {'category': 'physical', 'description': None},
                                  1: {'category': 'physical', 'description': 'r1 to r2', 'id': 'eth0'},
                                  2: {'category': 'physical', 'description': 'r1 to r3', 'id': 'eth1'},
                                  3: {'category': 'physical', 'description': 'r1 to r2', 'id': 'eth2'},
                                  4: {'category': 'physical', 'description': 'r1 to r3', 'id': 'eth3'},
                                  5: {'category': 'physical', 'description': 'r1 to r3', 'id': 'eth4'}},
                       'asn': 1,
                       'device_type': 'router',
                       'id': 'r1',
                       'label': 'r1',
                       'x': 350,
                       'y': 400},
                      {'_ports': {0: {'category': 'physical', 'description': None},
                                  1: {'category': 'physical', 'description': 'r2 to r1', 'id': 'eth0'},
                                  2: {'category': 'physical', 'description': 'r2 to r3', 'id': 'eth1'},
                                  3: {'category': 'physical', 'description': 'r2 to r4', 'id': 'eth2'},
                                  4: {'category': 'physical', 'description': 'r2 to r4', 'id': 'eth3'},
                                  5: {'category': 'physical', 'description': 'r2 to r1', 'id': 'eth4'}},
                       'asn': 1,
                       'device_type': 'router',
                       'id': 'r2',
                       'label': 'r2',
                       'x': 500,
                       'y': 300},
                      {'_ports': {0: {'category': 'physical', 'description': None},
                                  1: {'category': 'physical', 'description': 'r3 to r1', 'id': 'eth0'},
                                  2: {'category': 'physical', 'description': 'r3 to r2', 'id': 'eth1'},
                                  3: {'category': 'physical', 'description': 'r3 to r5', 'id': 'eth2'},
                                  4: {'category': 'physical', 'description': 'r3 to r1', 'id': 'eth3'},
                                  5: {'category': 'physical', 'description': 'r3 to r1', 'id': 'eth4'}},
                       'asn': 1,
                       'device_type': 'router',
                       'id': 'r3',
                       'label': 'r3',
                       'x': 500,
                       'y': 500}]}
    graph = json_graph.node_link_graph(data)
    anm = autonetkit.anm.NetworkModel()
    g_in = anm.add_overlay("input")
    g_in._replace_graph(nx.MultiGraph(graph))
    # TODO: check if should build overlays here rather than clone in?
    g_phy = anm["phy"]
    g_phy._replace_graph(graph)
    return anm



import autonetkit.log as log
from autonetkit.ank_utils import unwrap_edges, unwrap_nodes
from autonetkit.anm.base import OverlayBase
from autonetkit.anm.edge import NmEdge
from autonetkit.anm.interface import NmPort
from autonetkit.anm.node import NmNode
import autonetkit


class NmGraph(OverlayBase):

    """API to interact with an overlay graph in ANM"""

    @property
    def anm(self):
        """Returns anm for this overlay

        >>> anm = autonetkit.topos.house()

        """

        return self._anm

    @property
    def _graph(self):
        """Access underlying graph for this NmNode"""

        return self._anm.overlay_nx_graphs[self._overlay_id]

    def _replace_graph(self, graph):
        """"""

        self._anm.overlay_nx_graphs[self._overlay_id] = graph

    # these work similar to their nx counterparts: just need to strip the
    # node_id

    def _record_overlay_dependencies(self, nbunch):
        # TODO: add this logic to anm so can call when instantiating overlays too
        # TODO: make this able to be disabled for performance
        g_deps = self.anm['_dependencies']
        overlays = {n.overlay_id for n in nbunch if isinstance(n, NmNode)}
        if len(overlays) and self._overlay_id not in g_deps:
            g_deps.add_node(self._overlay_id)
        for overlay_id in overlays:
            if overlay_id not in g_deps:
                g_deps.add_node(overlay_id)

            if g_deps.number_of_edges(self._overlay_id, overlay_id) == 0:
                edge = (overlay_id, self._overlay_id)
                g_deps.add_edges_from([edge])

    def add_nodes_from( self, nbunch, retain=None, update=False,
        **kwargs):
        """Update won't append data (which could clobber) if node exists"""
        nbunch = list(nbunch)  # listify in case consumed in try/except

        if not retain:
            retain = []
        try:
            retain.lower()
            retain = [retain]  # was a string, put into list
        except AttributeError:
            pass  # already a list

        self._record_overlay_dependencies(nbunch)

        if not update:
            # TODO: what is update used for?
            # filter out existing nodes
            nbunch = (n for n in nbunch if n not in self._graph)

        nbunch = list(nbunch)
        node_ids = list(nbunch)  # before appending retain data

        if len(retain):
            add_nodes = []
            for node in nbunch:
                data = dict((key, node.get(key)) for key in retain)
                add_nodes.append((node.node_id, data))
            nbunch = add_nodes
        else:
            try:
                # only store the id in overlay
                nbunch = [n.node_id for n in nbunch]
            except AttributeError:
                pass  # use nbunch directly as the node IDs

        self._graph.add_nodes_from(nbunch, **kwargs)
        for node in self._graph.nodes():
            node_data = self._graph.node[node]
            if "label" not in node_data:
                # TODO: should this just fall through instead to phy/anm label
                # function?
                node_data["label"] = str(node)  # use node id

        self._copy_interfaces(node_ids)

    def _copy_interfaces(self, nbunch):
        """Copies ports from one overlay to another"""

        nbunch = [n.node_id if isinstance(n, NmNode) else n
                  for n in nbunch]

        if self._overlay_id == "phy":
            # note: this will auto copy data from input if present - by default
            # TODO: provide an option to add_nodes to skip this
            # or would it be better to just provide a function in ank_utils to
            # wipe interfaces in the rare case it's needed?
            input_graph = self._anm.overlay_nx_graphs['input']
            for node_id in nbunch:
                if node_id not in input_graph:
                    log.debug("Not copying interfaces for %s: ",
                              "not in input graph %s" % node_id)
                    self._graph.node[node_id]['_ports'] = {
                        0: {'description': 'loopback', 'category': 'loopback'}}
                    continue

                try:
                    input_interfaces = input_graph.node[node_id]['_ports']
                except KeyError:
                    #Node not in input
                    # Just do base initialisation of loopback zero
                    self._graph.node[node_id]['_ports'] = {
                        0: {'description': 'loopback', 'category': 'loopback'}}
                else:
                    interface_data = {'description': None,
                                      'category': 'physical'}
                    # need to do dict() to copy, otherwise all point to same memory
                    # location -> clobber
                    # TODO: update this to also get subinterfaces?
                    # TODO: should description and category auto fall through?
                    data = dict((key, dict(interface_data)) for key in
                                input_interfaces)
                    ports = {}
                    for key, vals in input_interfaces.items():
                        port_data = {}
                        ports[key] = dict(vals)

                    # force 0 to be loopback
                    # TODO: could warn if already set
                    ports[0] = {
                        'description': 'loopback', 'category': 'loopback'}
                    self._graph.node[node_id]['_ports'] = ports
            return

        if self._overlay_id == "graphics":
            # TODO: remove once graphics removed
            return

        phy_graph = self._anm.overlay_nx_graphs['phy']
        for node_id in nbunch:
            try:
                phy_interfaces = phy_graph.node[node_id]['_ports']
            except KeyError:
                # Node not in phy (eg broadcast domain)
                # Just do base initialisation of loopback zero
                self._graph.node[node_id]['_ports'] = {
                    0: {'description': 'loopback', 'category': 'loopback'}}
            else:
                interface_data = {'description': None,
                                  'category': 'physical'}
                # need to do dict() to copy, otherwise all point to same memory
                # location -> clobber
                # TODO: update this to also get subinterfaces?
                # TODO: should description and category auto fall through?
                data = dict((key, dict(interface_data)) for key in
                            phy_interfaces)
                self._graph.node[node_id]['_ports'] = data

    def add_node(self, node, retain=None, **kwargs):
        """Adds node to overlay"""
        nbunch = [node]
        self.add_nodes_from(nbunch,
                            retain, **kwargs)
        #TODO: test what this code is meant to do
        try:
            node_id = node.id
        except AttributeError:
            node_id = node  # use the string node id
        return NmNode(self.anm, self._overlay_id, node_id)

    def allocate_input_interfaces(self):
        """allocates edges to interfaces"""
        # TODO: move this to ank utils? or extra step in the anm?
        if self._overlay_id != "input":
            log.debug("Tried to allocate interfaces to %s" % overlay_id)
            return

        if all(len(node['input'].raw_interfaces) > 0 for node in self) \
            and all(len(edge['input'].raw_interfaces) > 0 for edge in
                    self.edges()):
            log.debug("Input interfaces allocated")
            return  # interfaces allocated
        elif self.data.interfaces_allocated == True:
            # explicitly flagged as allocated
            return
        else:
            log.info('Automatically assigning input interfaces')

        # Initialise loopback zero on node
        for node in self:
            node.raw_interfaces = {0:
                                   {'description': 'loopback', 'category': 'loopback'}}

        ebunch = sorted(self.edges())
        for edge in ebunch:
            src = edge.src
            dst = edge.dst
            src_int_id = src._add_interface('%s to %s' % (src.label,
                                                          dst.label))
            dst_int_id = dst._add_interface('%s to %s' % (dst.label,
                                                          src.label))
            edge.raw_interfaces = {
                src.id: src_int_id,
                dst.id: dst_int_id}

    def number_of_edges(self, node_a, node_b):
        return self._graph.number_of_edges(node_a, node_b)

    def __delitem__(self, key):
        """Alias for remove_node. Allows
        del overlay[node]
        """
        # TODO: needs to support node types
        self.remove_node(key)

    def remove_nodes_from(self, nbunch):
        """Removes set of nodes from nbunch"""

        try:
            nbunch = unwrap_nodes(nbunch)
        except AttributeError:
            pass  # don't need to unwrap

        self._graph.remove_nodes_from(nbunch)

    def remove_node(self, node_id):
        """Removes a node from the overlay"""
        if isinstance(node_id, NmNode):
            node_id = node_id.node_id

        self._graph.remove_node(node_id)

    def add_edge(self, src, dst, retain=None, **kwargs):
        """Adds an edge to the overlay"""

        if not retain:
            retain = []
        try:
            retain.lower()
            retain = [retain]  # was a string, put into list
        except AttributeError:
            pass  # already a list
        retval = self.add_edges_from([(src, dst)], retain, **kwargs)
        return retval[0]

    def remove_edges_from(self, ebunch):
        """Removes set of edges from ebunch"""

        try:
            ebunch = unwrap_edges(ebunch)
        except AttributeError:
            pass  # don't need to unwrap

        self._graph.remove_edges_from(ebunch)

    def add_edges(self, *args, **kwargs):
        """Adds a set of edges. Alias for add_edges_from"""

        self.add_edges_from(args, kwargs)

    def add_edges_from(self, ebunch, bidirectional=False, retain=None,
                       warn=True, **kwargs):
        """Add edges. Unlike NetworkX, can only add an edge if both
        src and dst in graph already.
        If they are not, then they will not be added (silently ignored)


        Retains interface mappings if they are present (this is why ANK
            stores the interface reference on the edges, as it simplifies
            cross-layer access, as well as split, aggregate, etc retaining the
            interface bindings)_

        Bidirectional will add edge in both directions. Useful if going
        from an undirected graph to a
        directed, eg G_in to G_bgp
        #TODO: explain "retain" and ["retain"] logic

        if user wants to add from another overlay, first go g_x.edges()
        then add from the result

        allow (src, dst, ekey), (src, dst, ekey, data) for the ank utils
        """

        if not retain:
            retain = []
        try:
            retain.lower()
            retain = [retain]  # was a string, put into list
        except AttributeError:
            pass  # already a list

        if self.is_multigraph():
            #used_keys = self._graph.adj[u][v]
            from collections import defaultdict
            used_keys = defaultdict(dict)

        all_edges = []
        for in_edge in ebunch:
            """Edge could be one of:
            - NmEdge
            - (NmNode, NmNode)
            - (NmPort, NmPort)
            - (NmNode, NmPort)
            - (NmPort, NmNode)
            - (string, string)
            """
            # This is less efficient than nx add_edges_from, but cleaner logic
            # TODO: could put the interface data into retain?
            data = {'_ports': {}}  # to retain
            ekey = None  # default is None (nx auto-allocates next int)

            # convert input to a NmEdge
            src = dst = None
            if isinstance(in_edge, NmEdge):
                edge = in_edge  # simple case
                ekey = edge.ekey  # explictly set ekey
                src = edge.src.node_id
                dst = edge.dst.node_id

                # and copy retain data
                data = dict((key, edge.get(key)) for key in retain)
                ports = {k: v for k, v in edge.raw_interfaces.items()
                         if k in self._graph}  # only if exists in this overlay
                # TODO: debug log if skipping a binding?
                data['_ports'] = ports

                # this is the only case where copy across data
                # but want to copy attributes for all cases

            elif len(in_edge) == 2:
                in_a, in_b = in_edge[0], in_edge[1]

                if isinstance(in_a, NmNode) and isinstance(in_b, NmNode):
                    src = in_a.node_id
                    dst = in_b.node_id

                elif isinstance(in_a, NmPort) and isinstance(in_b, NmPort):
                    src = in_a.node.node_id
                    dst = in_b.node.node_id
                    ports = {}
                    if src in self:
                        ports[src] = in_a.interface_id
                    if dst in self:
                        ports[dst] = in_b.interface_id
                    data['_ports'] = ports

                elif isinstance(in_a, NmNode) and isinstance(in_b, NmPort):
                    src = in_a.node_id
                    dst = in_b.node.node_id
                    ports = {}
                    if dst in self:
                        ports[dst] = in_b.interface_id
                    data['_ports'] = ports

                elif isinstance(in_a, NmPort) and isinstance(in_b, NmNode):
                    src = in_a.node.node_id
                    dst = in_b.node_id
                    ports = {}
                    if src in self:
                        ports[src] = in_a.interface_id
                    data['_ports'] = ports

                elif in_a in self and in_b in self:
                    src = in_a
                    dst = in_b

            elif len(in_edge) == 3:
                # (src, dst, ekey) format
                # or (src, dst, data) format
                in_a, in_b, in_c = in_edge[0], in_edge[1], in_edge[2]
                if in_a in self and in_b in self:
                    src = in_a
                    dst = in_b
                    # TODO: document the following logic
                    if self.is_multigraph() and not isinstance(in_c, dict):
                        ekey = in_c
                    else:
                        data = in_c

            elif len(in_edge) == 4:
                # (src, dst, ekey, data) format
                in_a, in_b = in_edge[0], in_edge[1]
                if in_a in self and in_b in self:
                    src = in_a
                    dst = in_b
                    ekey = in_edge[2]
                    data = in_edge[3]

            # TODO: if edge not set at this point, give error/warn

            # TODO: add check that edge.src and edge.dst exist
            if (src is None or dst is None) and warn:
                log.warning("Unsupported edge %s" % str(in_edge))
            if not(src in self and dst in self):
                if warn:
                    self.log.debug("Not adding edge %s, src/dst not in overlay"
                                   % str(in_edge))
                continue

            # TODO: warn if not multigraph and edge already exists - don't
            # add/clobber
            #TODO: double check this logic + add test case
            data.update(**kwargs)
            if self.is_multigraph() and ekey is None:
                # specifically allocate a key
                if src in used_keys and dst in used_keys[src]:
                    pass # already established
                else:
                    try:
                        used_keys[src][dst] = self._graph.adj[src][dst].keys()
                    except KeyError:
                        # no edges exist
                        used_keys[src][dst] = []

                # now have the keys mapping
                ekey=len(used_keys[src][dst])
                while ekey in used_keys[src][dst]:
                    ekey+=1

                used_keys[src][dst].append(ekey)

            edges_to_add = []
            if self.is_multigraph():
                edges_to_add.append((src, dst, ekey, dict(data)))
                if bidirectional:
                    edges_to_add.append((dst, src, ekey, dict(data)))
            else:
                edges_to_add.append((src, dst, dict(data)))
                if bidirectional:
                    edges_to_add.append((dst, src, dict(data)))


            #TODO: warn if not multigraph

            self._graph.add_edges_from(edges_to_add)
            all_edges += edges_to_add

        if self.is_multigraph():
            return [
            NmEdge(self.anm, self._overlay_id, src, dst, ekey) if ekey
            else NmEdge(self.anm, self._overlay_id, src, dst) # default no ekey set
            for src, dst, ekey, _ in all_edges]
        else:
            return [NmEdge(self.anm, self._overlay_id, src, dst)
            for src, dst, _ in all_edges]

    def update(self, nbunch=None, **kwargs):
        """Sets property defined in kwargs to all nodes in nbunch"""

        if nbunch is None:
            nbunch = self.nodes()

        if nbunch in self:
            nbunch = [nbunch]  # single node in the list for iteration

        # if the node is in the underlying networkx graph, then map to an
        # overlay node
        nbunch = [self.node(n) if n in self._graph else n for n in nbunch]
        for node in nbunch:
            for (key, value) in kwargs.items():
                node.set(key, value)

    def subgraph(self, nbunch, name=None):
        """"""

        nbunch = (n.node_id for n in nbunch)  # only store the id in overlay
        from autonetkit.anm.subgraph import OverlaySubgraph
        return OverlaySubgraph(self._anm, self._overlay_id,
                               self._graph.subgraph(nbunch), name)



"""
AutoNetkit Utilities
"""

#from anm import overlay_node, overlay_edge
import autonetkit


def call_log(fn, *args, **kwargs):
    def decorator(*args, **kwargs):
        # print "\t" + fn.__name__
        return fn(*args, **kwargs)

    return decorator


def unwrap_nodes(nodes):
    """Unwrap nodes"""
    from autonetkit.anm import NmNode

    try:
        return nodes.node_id  # treat as single node
    except AttributeError:
        if isinstance(nodes, basestring):
            return nodes  # string

        return [node.node_id if isinstance(node, NmNode)
                else node
                for node in nodes]  # treat as list


def unwrap_edges(edges):
    """Unwrap edges"""
    retval = []
    for edge in edges:
        if edge.is_multigraph():
            retval.append((edge.src_id, edge.dst_id, edge.ekey))
        else:
            retval.append((edge.src_id, edge.dst_id))

    return retval


def unwrap_graph(nm_graph):
    """Unwrap graph"""
    return nm_graph._graph


def alphabetical_sort(l):
    """From http://stackoverflow.com/questions/2669059/how-to-sort-alpha-numeric-set-in-python"""
# TODO: fix as currently only handles strings - not objects with repr?
    import re
    """ Sort the given iterable in the way that humans expect."""
    convert = lambda text: int(text) if text.isdigit() else text
    alphanum_key = lambda key: [convert(c) for c in re.split('([0-9]+)', key)]
    return sorted(l, key=alphanum_key)


def wrap_nodes(nm_graph, nodes):
    """ wraps node id into node overlay """
# TODO: remove duplicate of this in ank.py
    return (autonetkit.anm.overlay_node(nm_graph._anm, nm_graph._overlay_id, node)
            for node in nodes)


def merge_quagga_conf():
    # helper function to merge the quagga config files into a single file for
    # http://www.nongnu.org/quagga/docs/docs-multi/VTY-shell-integrated-configuration.html
    import pkg_resources
    import os
    import time

    templates_path = pkg_resources.resource_filename(__name__, "templates")
    zebra_dir = os.path.join(templates_path, "quagga", "etc", "zebra")
    # TODO: check the ordering - seems to matter
    conf_files = ["zebra.conf.mako", "bgpd.conf.mako",
                  "ospfd.conf.mako", "isisd.conf.mako"]
    timestamp = time.strftime('%Y%m%d_%H%M%S', time.localtime())
    data = ["##Merged by AutoNetkit merge_quagga_conf on %s" % timestamp]
    # TODO: replace newlines with !
    for conf_file in conf_files:
        filename = os.path.join(zebra_dir, conf_file)
        with open(filename) as fh:
            data.append(fh.read())

    ospfd_conf = os.path.join(templates_path, "quagga.conf.mako")
    with open(ospfd_conf, "w") as fh:
        fh.write("\n".join(data))


"""Potential edge utils:
# TODO: see if these are still used
def attr_equal(self, *args):
   ""Return edges which both src and dst have attributes equal""

    return all(getattr(self.src, key) == getattr(self.dst, key)
               for key in args)

def attr_both(self, *args):
    ""Return edges which both src and dst have attributes set""

    return all(getattr(self.src, key) and getattr(self.dst, key)
               for key in args)

def attr_any(self, *args):
    ""Return edges which either src and dst have attributes set""

    return all(getattr(self.src, key) or getattr(self.dst, key)
               for key in args)
"""
